from typing import Any, Dict, Iterable, List, Optional

import pandas as pd
import polars as pl
import pyarrow as pa
from triad.collections.schema import Schema
from triad.exceptions import InvalidOperationError
from triad.utils.assertion import assert_or_throw

from fugue import ArrowDataFrame
from fugue.api import (
    drop_columns,
    get_column_names,
    get_schema,
    is_df,
    rename,
    select_columns,
)
from fugue.dataframe.dataframe import DataFrame, LocalBoundedDataFrame, _input_schema
from fugue.dataset.api import (
    as_local,
    as_local_bounded,
    count,
    get_num_partitions,
    is_bounded,
    is_empty,
    is_local,
)
from fugue.exceptions import FugueDataFrameOperationError

from ._utils import build_empty_pl, pl_as_arrow, to_schema


class PolarsDataFrame(LocalBoundedDataFrame):
    """DataFrame that wraps :func:`pyarrow.Table <pa:pyarrow.table>`. Please also read
    |DataFrameTutorial| to understand this Fugue concept

    :param df: polars DataFrame or None, defaults to None
    :param schema: |SchemaLikeObject|
    """

    def __init__(
        self,
        df: Optional[pl.DataFrame] = None,
        schema: Any = None,
    ):
        if df is None:
            schema = _input_schema(schema).assert_not_empty()
            self._native: pa.Table = build_empty_pl(schema)
            super().__init__(schema)
            return
        else:
            assert_or_throw(
                schema is None,
                InvalidOperationError("can't reset schema for pl.DataFrame"),
            )
            self._native = df
            super().__init__(to_schema(df))

    @property
    def native(self) -> pl.DataFrame:
        """:func:`pyarrow.Table <pa:pyarrow.table>`"""
        return self._native

    def native_as_df(self) -> pl.DataFrame:
        return self._native

    @property
    def empty(self) -> bool:
        return self._native.shape[0] == 0

    def peek_array(self) -> List[Any]:
        self.assert_not_empty()
        return list(self._native.row(0))

    def peek_dict(self) -> Dict[str, Any]:
        self.assert_not_empty()
        return dict(zip(self._native.columns, self._native.row(0)))

    def count(self) -> int:
        return self.native.shape[0]

    def as_pandas(self) -> pd.DataFrame:
        return self.native.to_pandas()

    def head(
        self, n: int, columns: Optional[List[str]] = None
    ) -> LocalBoundedDataFrame:
        adf = self.native if columns is None else self.native.select(columns)
        n = min(n, self.count())
        if n == 0:
            schema = self.schema if columns is None else self.schema.extract(columns)
            return PolarsDataFrame(None, schema=schema)
        return PolarsDataFrame(adf.head(n))

    def _drop_cols(self, cols: List[str]) -> DataFrame:
        return PolarsDataFrame(self.native.drop(cols))

    def _select_cols(self, keys: List[Any]) -> DataFrame:
        return PolarsDataFrame(self.native.select(keys))

    def rename(self, columns: Dict[str, str]) -> DataFrame:
        return PolarsDataFrame(_rename_pl_dataframe(self.native, columns))

    def alter_columns(self, columns: Any) -> DataFrame:
        adf = ArrowDataFrame(self.as_arrow()).alter_columns(columns)
        return PolarsDataFrame(pl.from_arrow(adf.native))

    def as_arrow(self, type_safe: bool = False) -> pa.Table:
        return pl_as_arrow(self.native)

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        tdf = self.native
        if columns is not None:
            tdf = tdf.select(columns)
        return [list(row) for row in tdf.rows()]

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        if not self.empty:
            yield from ArrowDataFrame(pl_as_arrow(self.native)).as_array_iterable(
                columns=columns
            )

    def as_dict_iterable(
        self, columns: Optional[List[str]] = None
    ) -> Iterable[Dict[str, Any]]:
        if not self.empty:
            yield from ArrowDataFrame(pl_as_arrow(self.native)).as_dict_iterable(
                columns=columns
            )


@as_local.candidate(lambda df: isinstance(df, pl.DataFrame))
def _pl_as_local(df: pl.DataFrame) -> pl.DataFrame:
    return df


@as_local_bounded.candidate(lambda df: isinstance(df, pl.DataFrame))
def _pl_as_local_bounded(df: pl.DataFrame) -> pl.DataFrame:
    return df


@is_df.candidate(lambda df: isinstance(df, pl.DataFrame))
def _pl_is_df(df: pl.DataFrame) -> bool:
    return True


@count.candidate(lambda df: isinstance(df, pl.DataFrame))
def _pl_count(df: pl.DataFrame) -> int:
    return df.shape[0]


@is_bounded.candidate(lambda df: isinstance(df, pl.DataFrame))
def _pl_is_bounded(df: pl.DataFrame) -> bool:
    return True


@is_empty.candidate(lambda df: isinstance(df, pl.DataFrame))
def _pl_is_empty(df: pl.DataFrame) -> bool:
    return df.shape[0] == 0


@is_local.candidate(lambda df: isinstance(df, pl.DataFrame))
def _pl_is_local(df: pl.DataFrame) -> bool:
    return True


@get_num_partitions.candidate(lambda df: isinstance(df, pl.DataFrame))
def _pl_get_num_partitions(df: pl.DataFrame) -> int:
    return 1


@get_column_names.candidate(lambda df: isinstance(df, pl.DataFrame))
def _get_pl_columns(df: pl.DataFrame) -> List[Any]:
    return list(df.schema.keys())


@get_schema.candidate(lambda df: isinstance(df, pl.DataFrame))
def _get_pl_schema(df: pl.DataFrame) -> Schema:
    return to_schema(df)


@rename.candidate(lambda df, *args, **kwargs: isinstance(df, pl.DataFrame))
def _rename_pl_dataframe(df: pl.DataFrame, columns: Dict[str, Any]) -> pl.DataFrame:
    if len(columns) == 0:
        return df
    assert_or_throw(
        set(columns.keys()).issubset(set(df.columns)),
        FugueDataFrameOperationError(f"invalid {columns}"),
    )
    return df.rename(columns)


@drop_columns.candidate(lambda df, *args, **kwargs: isinstance(df, pl.DataFrame))
def _drop_pa_columns(df: pl.DataFrame, columns: List[str]) -> pl.DataFrame:
    cols = [x for x in df.schema.keys() if x not in columns]
    if len(cols) == 0:
        raise FugueDataFrameOperationError("cannot drop all columns")
    if len(cols) + len(columns) != len(df.columns):
        _assert_no_missing(df, columns)
    return df.select(cols)


@select_columns.candidate(lambda df, *args, **kwargs: isinstance(df, pl.DataFrame))
def _select_pa_columns(df: pl.DataFrame, columns: List[Any]) -> pl.DataFrame:
    if len(columns) == 0:
        raise FugueDataFrameOperationError("must select at least one column")
    _assert_no_missing(df, columns=columns)
    return df.select(columns)


def _assert_no_missing(df: pl.DataFrame, columns: Iterable[Any]) -> None:
    missing = [x for x in columns if x not in df.schema.keys()]
    if len(missing) > 0:
        raise FugueDataFrameOperationError("found nonexistent columns: {missing}")



from typing import Any, Dict

import ray

FUGUE_RAY_CONF_SHUFFLE_PARTITIONS = "fugue.ray.shuffle.partitions"
FUGUE_RAY_DEFAULT_PARTITIONS = "fugue.ray.default.partitions"
FUGUE_RAY_DEFAULT_BATCH_SIZE = "fugue.ray.default.batch_size"
FUGUE_RAY_ZERO_COPY = "fugue.ray.zero_copy"

FUGUE_RAY_DEFAULT_CONF: Dict[str, Any] = {
    FUGUE_RAY_CONF_SHUFFLE_PARTITIONS: -1,
    FUGUE_RAY_DEFAULT_PARTITIONS: 0,
    FUGUE_RAY_ZERO_COPY: True,
}

if ray.__version__ >= "2.3":
    _ZERO_COPY: Dict[str, Any] = {"zero_copy_batch": True}
else:  # pragma: no cover
    _ZERO_COPY = {}



import pickle

from fugue.rpc import make_rpc_server, to_rpc_handler, RPCFunc, EmptyRPCHandler
from pytest import raises
from triad import ParamDict


def test_default_server():
    def k(value: str) -> str:
        return value + "x"

    def kk(value: str) -> str:
        return value + "xx"

    conf = {"x": "y"}

    with make_rpc_server(conf).start() as server:
        assert "y" == server.conf["x"]
        with server.start():  # recursive start will take no effect
            client = server.make_client(k)
        assert "dddx" == client("ddd")
        client = server.make_client(kk)
        assert "dddxx" == client("ddd")
        server.stop()  # extra stop in the end will take no effect

    with raises(pickle.PicklingError):
        pickle.dumps(client)

    with raises(pickle.PicklingError):
        pickle.dumps(server)


def test_server_handlers():
    func = lambda x: x + "aa"

    class _Dict(RPCFunc):
        def __init__(self, obj):
            super().__init__(obj)
            self.start_called = 0
            self.stop_called = 0

        def start_handler(self):
            self.start_called += 1

        def stop_handler(self):
            self.stop_called += 1

    server = make_rpc_server({})
    server.start()
    d1 = _Dict(func)
    c1 = server.make_client(d1)
    assert "xaa" == c1("x")
    assert 1 == d1.start_called
    assert 0 == d1.stop_called
    server.stop()
    assert 1 == d1.start_called
    assert 1 == d1.stop_called

    with server.start():
        d2 = _Dict(func)
        c1 = server.make_client(d2)
        server.start()
        assert "xaa" == c1("x")
        assert 1 == d2.start_called
        assert 0 == d2.stop_called
        assert 1 == d1.start_called
        assert 1 == d1.stop_called
        server.stop()
    assert 1 == d2.start_called
    assert 1 == d2.stop_called
    assert 1 == d1.start_called
    assert 1 == d1.stop_called


def test_to_rpc_handler():
    assert isinstance(to_rpc_handler(None), EmptyRPCHandler)
    assert isinstance(to_rpc_handler(lambda x: x), RPCFunc)
    handler = to_rpc_handler(lambda x: x)
    assert handler is to_rpc_handler(handler)
    raises(ValueError, lambda: to_rpc_handler(1))



from typing import Dict, Optional

import dask.dataframe as pd
import numpy as np
import pandas
import pyarrow as pa
from dask.distributed import Client, get_client
from qpd_dask.engine import DaskUtils as DaskUtilsBase
from triad.utils.pyarrow import to_pandas_dtype, to_single_pandas_dtype

import fugue.api as fa
from fugue.constants import FUGUE_CONF_DEFAULT_PARTITIONS

from ._constants import FUGUE_DASK_CONF_DEFAULT_PARTITIONS


def get_default_partitions() -> int:
    conf = fa.get_current_conf()
    n = conf.get(
        FUGUE_DASK_CONF_DEFAULT_PARTITIONS,
        conf.get(FUGUE_CONF_DEFAULT_PARTITIONS, -1),
    )
    return n if n > 0 else fa.get_current_parallelism() * 2


class DaskUtils(DaskUtilsBase):
    def get_or_create_client(self, client: Optional[Client] = None):
        if client is not None:
            return client
        try:
            return get_client()
        except ValueError:
            return Client(processes=True)

    def is_compatile_index(self, df: pd.DataFrame) -> bool:
        """Check whether the datafame is compatible with the operations inside
        this utils collection

        :param df: dask dataframe
        :return: if it is compatible
        """
        if isinstance(df.index, pd.Index):
            return True
        return isinstance(df.index, pandas.RangeIndex) or (  # pragma: no cover
            hasattr(df.index, "inferred_type") and df.index.inferred_type == "integer"
        )

    def safe_to_pandas_dtype(self, schema: pa.Schema) -> Dict[str, np.dtype]:
        """Safely convert pyarrow types to pandas types. It will convert nested types
        to numpy object type. And this does not convert to pandas extension types.

        :param schema: the input pyarrow schema
        :return: the dictionary of numpy types

        .. note::

            This is a temporary solution, it will be removed when we use the Slide
            package. Do not use this function directly.
        """
        res: Dict[str, np.dtype] = {}
        for f in schema:
            if pa.types.is_nested(f.type):
                res[f.name] = np.dtype(object)
            else:
                res[f.name] = to_single_pandas_dtype(f.type, use_extension_types=False)
        return res

    # TODO: merge this back to base class
    def enforce_type(  # noqa: C901
        self, df: pd.DataFrame, schema: pa.Schema, null_safe: bool = False
    ) -> pd.DataFrame:  # pragma: no cover
        """Enforce the pandas like dataframe to comply with `schema`.
        :param df: pandas like dataframe
        :param schema: pyarrow schema
        :param null_safe: whether to enforce None value for int, string and bool values
        :return: converted dataframe
        :Notice:
        When `null_safe` is true, the native column types in the dataframe may change,
        for example, if a column of `int64` has None values, the output will make sure
        each value in the column is either None or an integer, however, due to the
        behavior of pandas like dataframes, the type of the columns may
        no longer be `int64`
        This method does not enforce struct and list types
        """
        if not null_safe:
            return df.astype(dtype=to_pandas_dtype(schema))
        for v in schema:
            s = df[v.name]
            if pa.types.is_string(v.type):
                ns = s.isnull()
                s = s.astype(str).mask(ns, None)
            elif pa.types.is_boolean(v.type):
                ns = s.isnull()
                if pandas.api.types.is_string_dtype(s.dtype):
                    try:
                        s = s.str.lower() == "true"
                    except AttributeError:
                        s = s.fillna(0).astype(bool)
                else:
                    s = s.fillna(0).astype(bool)
                s = s.mask(ns, None)
            elif pa.types.is_integer(v.type):
                ns = s.isnull()
                s = s.fillna(0).astype(v.type.to_pandas_dtype()).mask(ns, None)
            elif not pa.types.is_struct(v.type) and not pa.types.is_list(v.type):
                s = s.astype(v.type.to_pandas_dtype())
            df[v.name] = s
        return df


DASK_UTILS = DaskUtils()



# flake8: noqa
from fugue.extensions.processor.convert import (
    _to_processor,
    parse_processor,
    processor,
    register_processor,
)
from fugue.extensions.processor.processor import Processor



import pytest

ibis = pytest.importorskip("ibis")
import duckdb
from fugue import FugueWorkflow, NativeExecutionEngine
from fugue_ibis import IbisEngine, run_ibis
from fugue_test.ibis_suite import IbisTests

from fugue_duckdb import DuckDBIbisEngine, DuckExecutionEngine


class DuckDBIbisTests(IbisTests.Tests):
    @classmethod
    def setUpClass(cls):
        cls._con = duckdb.connect()
        cls._engine = cls.make_engine(cls)
        cls._ibis_engine = cls.make_ibis_engine(cls)

    @classmethod
    def tearDownClass(cls):
        cls._con.close()

    def make_engine(self):
        e = DuckExecutionEngine(dict(test=True), self._con)
        return e

    def make_ibis_engine(self) -> IbisEngine:
        return DuckDBIbisEngine(self._engine)

    def test_run_ibis_duck(self):
        def _test1(con: ibis.BaseBackend) -> ibis.Expr:
            tb = con.table("a")
            return tb

        def _test2(con: ibis.BaseBackend) -> ibis.Expr:
            tb = con.table("a")
            return tb.mutate(c=tb.a + tb.b)

        dag = FugueWorkflow()
        df = dag.df([[0, 1], [2, 3]], "a:long,b:long")
        res = run_ibis(_test1, ibis_engine="duck", a=df)
        res.assert_eq(df)
        df = dag.df([[0, 1], [2, 3]], "a:long,b:long")
        res = run_ibis(_test2, ibis_engine="duckdb", a=df)
        df2 = dag.df([[0, 1, 1], [2, 3, 5]], "a:long,b:long,c:long")
        res.assert_eq(df2)
        dag.run(self.engine)
        dag.run(NativeExecutionEngine())



from typing import Any

import pandas as pd
import pyarrow as pa
from fugue.dataframe import ArrowDataFrame
from fugue_test.dataframe_suite import DataFrameTests
from pytest import raises
import fugue.api as fa


class ArrowDataFrameTests(DataFrameTests.Tests):
    def df(self, data: Any = None, schema: Any = None) -> ArrowDataFrame:
        return ArrowDataFrame(data, schema)


class NativeArrowDataFrameTests(DataFrameTests.NativeTests):
    def df(self, data: Any = None, schema: Any = None) -> pd.DataFrame:
        return ArrowDataFrame(data, schema).as_arrow()

    def to_native_df(self, pdf: pd.DataFrame) -> Any:  # pragma: no cover
        return pa.Table.from_pandas(pdf)

    def test_num_partitions(self):
        assert fa.get_num_partitions(self.df([[0, 1]], "a:int,b:int")) == 1


def test_init():
    df = ArrowDataFrame(schema="a:str,b:int")
    assert df.empty
    assert df.schema == "a:str,b:int"
    assert df.is_bounded

    df = ArrowDataFrame(pd.DataFrame([], columns=["a", "b"]), schema="a:str,b:int")
    assert df.empty
    assert df.schema == "a:str,b:int"
    assert df.is_bounded

    data = [["a", "1"], ["b", "2"]]
    df = ArrowDataFrame(data, "a:str,b:str")
    assert [["a", "1"], ["b", "2"]] == df.as_array(type_safe=True)
    data = [["a", 1], ["b", 2]]
    df = ArrowDataFrame(data, "a:str,b:int")
    assert [["a", 1.0], ["b", 2.0]] == df.as_array(type_safe=True)
    df = ArrowDataFrame(data, "a:str,b:double")
    assert [["a", 1.0], ["b", 2.0]] == df.as_array(type_safe=True)

    ddf = ArrowDataFrame(df.native)
    assert [["a", 1.0], ["b", 2.0]] == ddf.as_array(type_safe=True)

    df = ArrowDataFrame(df.as_pandas(), "a:str,b:double")
    assert [["a", 1.0], ["b", 2.0]] == df.as_array(type_safe=True)
    df = ArrowDataFrame(df.as_pandas()["b"])
    assert [[1.0], [2.0]] == df.as_array(type_safe=True)

    df = ArrowDataFrame([], "x:str,y:double")
    assert df.empty
    assert df.is_local
    assert df.is_bounded

    raises(Exception, lambda: ArrowDataFrame(123))



import json

from fugue.collections.partition import parse_presort_exp, PartitionSpec
from fugue.constants import KEYWORD_PARALLELISM, KEYWORD_ROWCOUNT
from pytest import raises
from triad.collections.schema import Schema
from triad.utils.hash import to_uuid
from triad.collections.dict import IndexedOrderedDict


def test_parse_presort_exp():

    assert parse_presort_exp(None) == IndexedOrderedDict()
    assert parse_presort_exp(IndexedOrderedDict([("c", True)])) == IndexedOrderedDict(
        [("c", True)]
    )
    assert parse_presort_exp("c") == IndexedOrderedDict([("c", True)])
    assert parse_presort_exp("         c") == IndexedOrderedDict([("c", True)])
    assert parse_presort_exp("c           desc") == IndexedOrderedDict([("c", False)])
    assert parse_presort_exp("b desc, c asc") == IndexedOrderedDict(
        [("b", False), ("c", True)]
    )
    assert parse_presort_exp("DESC DESC, ASC ASC") == IndexedOrderedDict(
        [("DESC", False), ("ASC", True)]
    )
    assert parse_presort_exp([("b", False), ("c", True)]) == IndexedOrderedDict(
        [("b", False), ("c", True)]
    )
    assert parse_presort_exp("B DESC, C ASC") == IndexedOrderedDict(
        [("B", False), ("C", True)]
    )
    assert parse_presort_exp("b desc, c asc") == IndexedOrderedDict(
        [("b", False), ("c", True)]
    )
    assert parse_presort_exp("`` desc, `a b` asc, ````, `中国`") == IndexedOrderedDict(
        [("", False), ("a b", True), ("`", True), ("中国", True)]
    )
    assert parse_presort_exp([("", False), ("a b", True), "中国"]) == IndexedOrderedDict(
        [("", False), ("a b", True), ("中国", True)]
    )

    with raises(SyntaxError):
        parse_presort_exp("b dsc, c asc")  # mispelling of desc

    with raises(SyntaxError):
        parse_presort_exp("c true")  # string format needs desc/asc

    with raises(SyntaxError):
        parse_presort_exp("c true, c true")  # cannot contain duplicates

    with raises(SyntaxError):
        parse_presort_exp("a b dsc, c asc")  # no quote

    with raises(SyntaxError):
        parse_presort_exp(
            [("b", "desc"), ("c", "asc")]
        )  # instead of desc and asc, needs to be bool


def test_partition_spec():
    p = PartitionSpec()
    assert [] == p.partition_by
    "0" == p.num_partitions
    {} == p.presort
    "hash" == p.algo
    assert p.empty

    p = PartitionSpec(None)
    assert p.empty
    p2 = PartitionSpec(p)
    assert p2.empty

    p = PartitionSpec(json.dumps(dict(partition_by=["a", "b", "c"], num_partitions=1)))
    assert ["a", "b", "c"] == p.partition_by
    assert "1" == p.num_partitions
    assert {} == p.presort
    assert "hash" == p.algo
    assert not p.empty

    p = PartitionSpec(dict(by=["a", "b", "c"], presort="d asc,e desc"))
    assert ["a", "b", "c"] == p.partition_by
    assert "0" == p.num_partitions
    assert dict(d=True, e=False) == p.presort
    assert "hash" == p.algo
    assert not p.empty

    p = PartitionSpec(by=["a ", "b", "c"], num=5, presort="d,`e ` desc", algo="EvEN")
    assert ["a ", "b", "c"] == p.partition_by
    assert "5" == p.num_partitions
    assert {"d": True, "e ": False} == p.presort
    assert "even" == p.algo
    assert not p.empty

    p = PartitionSpec(
        partition_by=["a", "b", "c"],
        presort="d,e desc",
        algo="EvEN",
        num_partitions="ROWCOUNT*3",
        row_limit=4,
        size_limit="5k",
    )
    p2 = PartitionSpec(p)
    assert p2.jsondict == p.jsondict
    assert "d ASC,e DESC" == p2.presort_expr
    assert not p.empty
    assert not p2.empty
    print(p)
    print(f"{p}")

    assert PartitionSpec("per_row") == PartitionSpec(num="ROWCOUNT", algo="even")
    assert PartitionSpec(by="abc") == PartitionSpec(by=["abc"])
    assert PartitionSpec("abc") == PartitionSpec(by=["abc"])
    assert PartitionSpec(["abc"]) == PartitionSpec(by=["abc"])
    assert PartitionSpec(["abc", "def"]) == PartitionSpec(by=["abc", "def"])
    assert PartitionSpec(("abc", "def")) == PartitionSpec(by=["abc", "def"])

    assert PartitionSpec(4) == PartitionSpec(num=4)

    # partition by overlaps with presort
    raises(
        SyntaxError,
        lambda: PartitionSpec(
            partition_by=["a", "b", "c"], presort="a asc,e desc", algo="EvEN"
        ),
    )

    # partition by has dups
    raises(SyntaxError, lambda: PartitionSpec(partition_by=["a", "b", "b"]))

    # partition by has dups
    raises(SyntaxError, lambda: PartitionSpec(partition_by=["a", "b", "b"]))

    # partition by is invalid
    raises(SyntaxError, lambda: PartitionSpec(partition_by=123))

    # bad input
    raises(TypeError, lambda: PartitionSpec(1.1))

    # bad presort
    raises(SyntaxError, lambda: PartitionSpec(presort="a xsc,e desc"))
    raises(SyntaxError, lambda: PartitionSpec(presort="a asc,a desc"))
    raises(SyntaxError, lambda: PartitionSpec(presort="a b asc,a desc"))
    raises(SyntaxError, lambda: PartitionSpec(presort=[("a", "asc"), "b"]))
    raises(SyntaxError, lambda: PartitionSpec(presort=[("a",), ("b")]))
    raises(SyntaxError, lambda: PartitionSpec(presort=["a", ["b", True]]))

    p = PartitionSpec(dict(partition_by=["a"], presort="d asc,e desc"))
    assert dict(a=True, d=True, e=False) == p.get_sorts(
        Schema("a:int,b:int,d:int,e:int")
    )
    assert dict(d=True, e=False) == p.get_sorts(
        Schema("a:int,b:int,d:int,e:int"), with_partition_keys=False
    )
    p = PartitionSpec(dict(partition_by=["e", "a"], presort="d asc"))
    assert p.get_key_schema(Schema("a:int,b:int,d:int,e:int")) == "e:int,a:int"

    # modification
    a = PartitionSpec(by=["a", "b"])
    b = PartitionSpec(a, by=["a"], num=2)
    assert ["a", "b"] == a.partition_by
    assert "0" == a.num_partitions
    assert ["a"] == b.partition_by
    assert "2" == b.num_partitions

    a = PartitionSpec(by=["a"], presort="b DESC, c")
    b = PartitionSpec(by=["a"], presort="c,b DESC")
    assert a.presort != b.presort
    c = PartitionSpec(b, presort=a.presort)
    assert a.presort == c.presort
    c = PartitionSpec(b, presort=[("b", False), ("c", True)])
    assert a.presort == c.presort

    a = PartitionSpec(by=["a"], presort="b DESC, c")
    b = PartitionSpec(by=["a"], presort=[("c", True), ("b", False)])
    c = PartitionSpec(by=["a"], presort=["c", ("b", False)])
    d = PartitionSpec(by=["a"], presort=[("b", False), ("c", True)])
    assert a.presort != b.presort
    assert b.presort == c.presort
    assert a.presort == d.presort

    a = PartitionSpec(by=["a"], presort=["b", "c"])
    b = PartitionSpec(by=["a"], presort=[("b"), ("c")])
    c = PartitionSpec(by=["a"], presort=[("b", True), ("c", True)])
    d = PartitionSpec(by=["a"], presort="b asc, c")
    assert a.presort == b.presort
    assert a.presort == c.presort
    assert a.presort == d.presort

    # test eq
    a = PartitionSpec(by=["a"], presort=["b", "c"])
    b = PartitionSpec(presort="b, c asc", by=["a"])
    c = PartitionSpec(num=10, by=["a"], presort=["b", "c"])
    d = PartitionSpec(num=10, by=["a"], presort=["c", "b"])
    assert a == a
    assert a == dict(presort=["b", "c"], by=["a"])
    assert a == b
    assert a != c
    assert c != d


def test_partition_cursor():
    p = PartitionSpec(dict(partition_by=["b", "a"]))
    s = Schema("a:int,b:int,c:int,d:int")
    c = p.get_cursor(s, 2)
    pt = p.get_partitioner(s)  # this part is well covered in spark section
    assert c.row_schema == s
    assert c.key_schema == "b:int,a:int"

    c.set([1, 2, 2, 2], 5, 6)
    assert [2, 1] == c.key_value_array
    assert dict(a=1, b=2) == c.key_value_dict
    assert 2 == c["c"]
    assert [1, 2, 2, 2] == c.row
    assert 5 == c.partition_no
    assert 2 == c.physical_partition_no
    assert 6 == c.slice_no

    c.set(lambda: [1, 2, 2, 2], 5, 6)
    assert [2, 1] == c.key_value_array
    assert dict(a=1, b=2) == c.key_value_dict


def test_get_num_partitions():
    p = PartitionSpec(dict(partition_by=["b", "a"]))
    assert 0 == p.get_num_partitions()

    p = PartitionSpec(dict(partition_by=["b", "a"], num=123))
    assert 123 == p.get_num_partitions()

    p = PartitionSpec(dict(partition_by=["b", "a"], num="(x + Y) * 2"))
    assert 6 == p.get_num_partitions(x=lambda: 1, Y=lambda: 2)
    raises(Exception, lambda: p.get_num_partitions(x=lambda: 1))

    p = PartitionSpec(dict(partition_by=["b", "a"], num="min(ROWCOUNT,CONCURRENCY)"))
    assert 90 == p.get_num_partitions(
        **{KEYWORD_ROWCOUNT: lambda: 100, KEYWORD_PARALLELISM: lambda: 90}
    )


def test_determinism():
    a = PartitionSpec(num=0)
    b = PartitionSpec()
    assert to_uuid(a) == to_uuid(b)

    a = PartitionSpec(by=["a"], num=2)
    b = PartitionSpec(num="2", by=["a"])
    assert to_uuid(a) == to_uuid(b)

    a = PartitionSpec(by=["a", "b"])
    b = PartitionSpec(by=["b", "a"])
    assert to_uuid(a) != to_uuid(b)



# pylint: disable-all
# flake8: noqa
# TODO: This folder is to be deprecated
import warnings
from fugue.exceptions import *

warnings.warn(
    "fsql and FugueSQLWorkflow now should be imported directly from fugue, "
    "fugue_sql will be removed in 0.9.0"
)



from datetime import datetime
from typing import Iterable, Iterator

import polars as pl
import ray
from dask.distributed import Client
from pyspark.sql import SparkSession
import pandas as pd

import fugue.api as fa


def test_transform_common():
    def tr1(df: pl.DataFrame) -> pl.DataFrame:
        tdf = df.with_column(pl.lit(1, pl.Int32()).alias("b"))
        return tdf

    def tr2(dfs: Iterable[pl.DataFrame]) -> Iterator[pl.DataFrame]:
        for df in dfs:
            tdf = df.with_column(pl.lit(1, pl.Int32()).alias("b"))
            yield tdf

    for tr in [tr1, tr2]:
        df = fa.as_fugue_df([[0], [1]], schema="a:int")
        fdf = fa.transform(df, tr, schema="a:int,b:int", as_fugue=True)
        assert fdf.schema == "a:int,b:int"
        assert fdf.as_array() == [[0, 1], [1, 1]]

        df = fa.as_fugue_df([[0], [1]], schema="a:int")
        fdf = fa.transform(df, tr, schema="b:int,a:str", as_fugue=True)
        assert fdf.schema == "b:int,a:str"
        assert fdf.as_array() == [[1, "0"], [1, "1"]]

        # polars use large_string, so this needs to be handled
        df = fa.as_fugue_df([["0"], ["1"]], schema="a:str")
        fdf = fa.transform(df, tr, schema="b:int,a:str", as_fugue=True)
        assert fdf.schema == "b:int,a:str"
        assert fdf.as_array() == [[1, "0"], [1, "1"]]

        df = fa.as_fugue_df([], schema="a:int")
        fdf = fa.transform(df, tr, schema="a:int,b:int", as_fugue=True)
        assert fdf.schema == "a:int,b:int"
        assert fdf.as_array() == []

        df = pl.from_pandas(pd.DataFrame({"a": [0, 1]}))
        fdf = fa.transform(df, tr, schema="a:int,b:int", as_fugue=True)
        assert fdf.schema == "a:int,b:int"
        assert fdf.as_array() == [[0, 1], [1, 1]]


def test_transform_empty_result():
    def tr1(df: pl.DataFrame) -> pl.DataFrame:
        tdf = df.with_column(pl.lit(1, pl.Int32()).alias("b"))
        return tdf.head(0)

    def tr2(dfs: Iterable[pl.DataFrame]) -> Iterator[pl.DataFrame]:
        for _ in []:
            yield None

    def tr3(dfs: Iterable[pl.DataFrame]) -> Iterator[pl.DataFrame]:
        return

    for tr in [tr1, tr2, tr3]:
        df = fa.as_fugue_df([[0], [1]], schema="a:int")
        fdf = fa.transform(df, tr, schema="a:int,b:int", as_fugue=True)
        assert fdf.schema == "a:int,b:int"
        assert fdf.as_array() == []


def test_polars_on_engines():
    def tr1(df: pl.DataFrame) -> pl.DataFrame:
        tdf = df.with_column(pl.lit(1, pl.Int32()).alias("c"))
        return tdf

    def tr2(dfs: Iterable[pl.DataFrame]) -> Iterator[pl.DataFrame]:
        for df in dfs:
            tdf = df.with_column(pl.lit(1, pl.Int32()).alias("c"))
            yield tdf

    def test(engine):
        for tr in [tr1, tr2]:
            with fa.engine_context(engine):
                df = fa.as_fugue_df(
                    [["a", datetime(2022, 1, 1)], ["b", datetime(2022, 1, 2)]],
                    schema="a:str,b:datetime",
                )
                fdf = fa.transform(df, tr, schema="*,c:int", as_fugue=True)
                assert [
                    ["a", datetime(2022, 1, 1), 1],
                    ["b", datetime(2022, 1, 2), 1],
                ] == sorted(fdf.as_array(), key=lambda x: x[0])

    with SparkSession.builder.getOrCreate() as spark:
        test(spark)

    with Client(processes=True) as client:
        test(client)

    with ray.init():
        test("ray")



import copy
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional

import pandas as pd
import pyarrow as pa
from pytest import raises
from triad import to_uuid
from triad.utils.iter import EmptyAwareIterable

from fugue import (
    ArrayDataFrame,
    ArrowDataFrame,
    DataFrame,
    IterableDataFrame,
    LocalDataFrame,
    IterablePandasDataFrame,
    PandasDataFrame,
)
from fugue.dataframe.function_wrapper import _IterablePandasParam, _IterableArrowParam
from fugue.dataframe.utils import _df_eq as df_eq
from fugue.dev import DataFrameFunctionWrapper


def test_function_wrapper():
    for f in [f20, f21, f212, f22, f23, f24, f25, f26, f30, f31, f32, f35, f36]:
        df = ArrayDataFrame([[0]], "a:int")
        w = DataFrameFunctionWrapper(f, "^[ldsp][ldsp]$", "[ldspq]")
        res = w.run([df], dict(a=df), ignore_unknown=False, output_schema="a:int")
        df_eq(res, [[0], [0]], "a:int", throw=True)
        w.run([df], dict(a=df), ignore_unknown=False, output=False)

    # test other data types, simple operations
    w = DataFrameFunctionWrapper(f27)
    assert w.get_format_hint() is None
    assert 3 == w(1, 2)
    assert 3 == w.run([1, 2], dict(), ignore_unknown=False)
    assert 3 == w.run([5], dict(a=1, b=2), ignore_unknown=True)  # dict will overwrite
    assert 3 == w.run([], dict(a=1, b=2, c=4), ignore_unknown=True)
    raises(ValueError, lambda: w.run([], dict(a=1, b=2, c=4), ignore_unknown=False))

    # test default and required
    w = DataFrameFunctionWrapper(f28)
    assert 3 == w.run([], dict(a=1, b=2), ignore_unknown=False)
    assert 2 == w.run([], dict(a=1), ignore_unknown=False)
    assert 3 == w.run([], dict(a=1, b=2), ignore_unknown=True)
    assert 3 == w.run([], dict(a=1, b=2, c=4), ignore_unknown=True)
    raises(ValueError, lambda: w.run([], dict(a=1, b=2, c=4), ignore_unknown=False))
    raises(ValueError, lambda: w.run([], dict(b=2), ignore_unknown=True))

    # test kwargs
    w = DataFrameFunctionWrapper(f29)
    assert 3 == w.run([], dict(a=1, b=2), ignore_unknown=False)
    assert 1 == w.run([], dict(a=1), ignore_unknown=False)
    assert 3 == w.run([], dict(a=1, b=2), ignore_unknown=True)
    assert 7 == w.run([], dict(a=1, b=2, c=4), ignore_unknown=True)
    assert 7 == w.run([], dict(a=1, b=2, c=4), ignore_unknown=False)

    # test method inside class
    class Test(object):
        def t(self, a=1, b=2) -> int:
            return a + b

    test = Test()
    # instance method test
    w = DataFrameFunctionWrapper(test.t, "^0?.*", ".*")
    assert 4 == w.run([], kwargs={"b": 3}, ignore_unknown=True)
    assert 5 == w.run([2], kwargs={"b": 3}, ignore_unknown=True)

    # format hint
    w = DataFrameFunctionWrapper(f10)
    assert w.get_format_hint() == "pyarrow"

    w = DataFrameFunctionWrapper(f11)
    assert w.get_format_hint() == "pandas"

    w = DataFrameFunctionWrapper(f12)
    assert w.get_format_hint() == "pyarrow"

    w = DataFrameFunctionWrapper(f13)
    assert w.get_format_hint() == "pandas"

    w = DataFrameFunctionWrapper(f14)
    assert w.get_format_hint() == "pandas"


def test_function_wrapper_determinism():
    w1 = DataFrameFunctionWrapper(f20, "^[ldsp][ldsp]$", "[ldsp]")
    w2 = DataFrameFunctionWrapper(f20, "^[ldsp][ldsp]$", "[ldsp]")
    assert w1 is not w2
    assert to_uuid(w1) == to_uuid(w2)


def test_function_wrapper_copy():
    class Test(object):
        def __init__(self):
            self.n = 0

        def t(self) -> None:
            self.n += 1

    test = Test()
    w1 = DataFrameFunctionWrapper(test.t, "", "n")
    w2 = copy.copy(w1)
    w3 = copy.deepcopy(w1)
    w1.run([], {}, output=False)
    w2.run([], {}, output=False)
    w3.run([], {}, output=False)
    assert 3 == test.n


def test_iterable_pandas_dataframes():
    p = _IterablePandasParam(None)
    pdf = pd.DataFrame([[0]], columns=["a"])
    df = PandasDataFrame(pdf)
    data = list(p.to_input_data(df, ctx=None))
    assert 1 == len(data)
    assert data[0] is pdf  # this is to guarantee no copy in any wrapping logic
    assert data[0].values.tolist() == [[0]]

    dfs = IterablePandasDataFrame([df, df])
    data = list(p.to_input_data(dfs, ctx=None))
    assert 2 == len(data)
    assert data[0] is pdf
    assert data[1] is pdf

    def get_pdfs():
        yield pdf
        yield pdf

    # without schema change, there is no copy
    odf = p.to_output_df(get_pdfs(), df.schema, ctx=None)
    data = list(odf.native)
    assert 2 == len(data)
    assert data[0].native is pdf
    assert data[1].native is pdf

    # with schema change, there is copy
    odf = p.to_output_df(get_pdfs(), "a:double", ctx=None)
    data = list(odf.native)
    assert 2 == len(data)
    assert data[0].native is not pdf
    assert data[1].native is not pdf


def test_iterable_arrow_dataframes():
    p = _IterableArrowParam(None)
    pdf = pa.Table.from_pandas(pd.DataFrame([[0]], columns=["a"]))
    df = ArrowDataFrame(pdf)
    data = list(p.to_input_data(df, ctx=None))
    assert 1 == len(data)
    assert data[0] is pdf  # this is to guarantee no copy in any wrapping logic

    dfs = IterablePandasDataFrame([df, df])
    data = list(p.to_input_data(dfs, ctx=None))
    assert 2 == len(data)
    assert data[0] is pdf
    assert data[1] is pdf

    def get_pdfs():
        yield pdf
        yield pdf

    # without schema change, there is no copy
    odf = p.to_output_df(get_pdfs(), df.schema, ctx=None)
    data = list(odf.native)
    assert 2 == len(data)
    assert data[0].native is pdf
    assert data[1].native is pdf

    # with schema change, there is copy
    odf = p.to_output_df(get_pdfs(), "a:double", ctx=None)
    data = list(odf.native)
    assert 2 == len(data)
    assert data[0].native is not pdf
    assert data[1].native is not pdf


def f10(x: Any, y: pa.Table) -> None:
    pass


def f11(x: Any, y: pd.DataFrame, z: pa.Table) -> pa.Table:
    pass


def f12(x: Any, y) -> pa.Table:
    pass


def f13(x: Any, y: Iterable[pd.DataFrame]) -> pa.Table:
    pass


def f14() -> Iterator[pd.DataFrame]:
    pass


def f20(e: List[List[Any]], a: Iterable[List[Any]]) -> LocalDataFrame:
    e += list(a)
    return IterableDataFrame(e, "a:int")


def f21(e: List[Dict[str, Any]], a: Iterator[Dict[str, Any]]) -> DataFrame:
    e += list(a)
    arr = [[x["a"]] for x in e]
    return IterableDataFrame(arr, "a:int")


def f212(e: List[Dict[str, Any]], a: Iterable[Dict[str, Any]]) -> DataFrame:
    e += list(a)
    arr = [[x["a"]] for x in e]
    return ArrayDataFrame(arr, "a:int")


def f22(e: List[List[Any]], a: Iterator[List[Any]]) -> List[List[Any]]:
    e += list(a)
    return ArrayDataFrame(e, "a:int").as_array()


def f23(e: List[Dict[str, Any]], a: Iterable[Dict[str, Any]]) -> Iterable[List[Any]]:
    e += list(a)
    arr = [[x["a"]] for x in e]
    return ArrayDataFrame(arr, "a:int").as_array_iterable()


def f24(e: List[Dict[str, Any]], a: Iterable[Dict[str, Any]]) -> pd.DataFrame:
    e += list(a)
    arr = [[x["a"]] for x in e]
    return ArrayDataFrame(arr, "a:int").as_pandas()


def f25(e: DataFrame, a: LocalDataFrame) -> List[Dict[str, Any]]:
    e = e.as_array()
    e += list(a.as_array())
    return list(ArrayDataFrame(e, "a:int").as_dict_iterable())


def f26(e: pd.DataFrame, a: LocalDataFrame) -> Iterable[Dict[str, Any]]:
    e = list(PandasDataFrame(e).as_array())
    e += list(a.as_array())
    return ArrayDataFrame(e, "a:int").as_dict_iterable()


def f27(a, b):
    return a + b


def f28(a, b=1):
    return a + b


def f29(a, **args):
    for v in args.values():
        a += v
    return a


def f30(
    e: EmptyAwareIterable[List[Any]], a: EmptyAwareIterable[Dict[str, Any]]
) -> LocalDataFrame:
    e.peek()
    a.peek()
    e = list(e)
    e += [[x["a"]] for x in a]
    return ArrayDataFrame(e, "a:int")


def f31(
    e: List[Dict[str, Any]], a: Iterable[Dict[str, Any]]
) -> EmptyAwareIterable[List[Any]]:
    e += list(a)
    arr = [[x["a"]] for x in e]
    return ArrayDataFrame(arr, "a:int").as_array_iterable()


def f32(
    e: List[Dict[str, Any]], a: Iterable[Dict[str, Any]]
) -> EmptyAwareIterable[Dict[str, Any]]:
    e += list(a)
    arr = [[x["a"]] for x in e]
    return ArrayDataFrame(arr, "a:int").as_dict_iterable()


def f35(e: pd.DataFrame, a: LocalDataFrame) -> Iterable[pd.DataFrame]:
    e = PandasDataFrame(e, "a:int").as_pandas()
    a = ArrayDataFrame(a, "a:int").as_pandas()
    return iter([e, a])


def f36(e: pd.DataFrame, a: LocalDataFrame) -> Iterable[pa.Table]:
    e = PandasDataFrame(e, "a:int").as_arrow()
    a = ArrayDataFrame(a, "a:int").as_arrow()
    return iter([e, a])



import logging
import os
from typing import Any, Callable, Dict, List, Optional, Type, Union

import dask.dataframe as dd
import pandas as pd
from distributed import Client
from qpd_dask import run_sql_on_dask
from triad.collections import Schema
from triad.collections.dict import IndexedOrderedDict, ParamDict
from triad.collections.fs import FileSystem
from triad.utils.assertion import assert_or_throw
from triad.utils.hash import to_uuid
from triad.utils.threading import RunOnce

from fugue import StructuredRawSQL
from fugue.collections.partition import (
    PartitionCursor,
    PartitionSpec,
    parse_presort_exp,
)
from fugue.constants import KEYWORD_PARALLELISM, KEYWORD_ROWCOUNT
from fugue.dataframe import (
    AnyDataFrame,
    DataFrame,
    DataFrames,
    LocalDataFrame,
    PandasDataFrame,
)
from fugue.dataframe.utils import get_join_schemas
from fugue.execution.execution_engine import ExecutionEngine, MapEngine, SQLEngine
from fugue.execution.native_execution_engine import NativeExecutionEngine
from fugue_dask._constants import FUGUE_DASK_DEFAULT_CONF
from fugue_dask._io import load_df, save_df
from fugue_dask._utils import DASK_UTILS, DaskUtils
from fugue_dask.dataframe import DaskDataFrame

_DASK_PARTITION_KEY = "__dask_partition_key__"


class QPDDaskEngine(SQLEngine):
    """QPD execution implementation."""

    @property
    def dialect(self) -> Optional[str]:
        return "spark"

    def to_df(self, df: AnyDataFrame, schema: Any = None) -> DataFrame:
        return to_dask_engine_df(df, schema)

    @property
    def is_distributed(self) -> bool:
        return True

    def select(self, dfs: DataFrames, statement: StructuredRawSQL) -> DataFrame:
        _dfs, _sql = self.encode(dfs, statement)
        dask_dfs = {k: self.to_df(v).native for k, v in _dfs.items()}  # type: ignore
        df = run_sql_on_dask(_sql, dask_dfs, ignore_case=True)
        return DaskDataFrame(df)


class DaskMapEngine(MapEngine):
    @property
    def execution_engine_constraint(self) -> Type[ExecutionEngine]:
        return DaskExecutionEngine

    @property
    def is_distributed(self) -> bool:
        return True

    def map_dataframe(
        self,
        df: DataFrame,
        map_func: Callable[[PartitionCursor, LocalDataFrame], LocalDataFrame],
        output_schema: Any,
        partition_spec: PartitionSpec,
        on_init: Optional[Callable[[int, DataFrame], Any]] = None,
        map_func_format_hint: Optional[str] = None,
    ) -> DataFrame:
        is_coarse = partition_spec.algo == "coarse"
        presort = partition_spec.get_sorts(df.schema, with_partition_keys=is_coarse)
        presort_keys = list(presort.keys())
        presort_asc = list(presort.values())
        output_schema = Schema(output_schema)
        input_schema = df.schema
        cursor = partition_spec.get_cursor(input_schema, 0)
        on_init_once: Any = (
            None
            if on_init is None
            else RunOnce(
                on_init, lambda *args, **kwargs: to_uuid(id(on_init), id(args[0]))
            )
        )

        def _map(pdf: Any) -> pd.DataFrame:
            if pdf.shape[0] == 0:
                return PandasDataFrame([], output_schema).as_pandas()
            if is_coarse:
                pdf = pdf.drop(columns=[_DASK_PARTITION_KEY])
            if len(partition_spec.presort) > 0:
                pdf = pdf.sort_values(presort_keys, ascending=presort_asc)
            input_df = PandasDataFrame(
                pdf.reset_index(drop=True), input_schema, pandas_df_wrapper=True
            )
            if on_init_once is not None:
                on_init_once(0, input_df)
            cursor.set(lambda: input_df.peek_array(), 0, 0)
            output_df = map_func(cursor, input_df)
            return output_df.as_pandas()[output_schema.names]

        df = self.to_df(df)
        meta = self.execution_engine.pl_utils.safe_to_pandas_dtype(  # type: ignore
            output_schema.pa_schema
        )
        if len(partition_spec.partition_by) == 0:
            pdf = self.execution_engine.repartition(df, partition_spec)
            result = pdf.native.map_partitions(_map, meta=meta)  # type: ignore
        else:
            df = self.execution_engine.repartition(
                df, PartitionSpec(num=partition_spec.num_partitions)
            )
            if is_coarse:
                input_num_partitions = df.num_partitions
                _utils = self.execution_engine.pl_utils  # type: ignore
                input_meta = _utils.safe_to_pandas_dtype(
                    (input_schema + (_DASK_PARTITION_KEY, "uint64")).pa_schema
                )
                tddf = df.native.map_partitions(
                    lambda pdf: pdf.assign(
                        **{
                            _DASK_PARTITION_KEY: pd.util.hash_pandas_object(
                                pdf[partition_spec.partition_by], index=False
                            ).mod(input_num_partitions)
                        }
                    ),
                    meta=input_meta,
                )
                keys = [_DASK_PARTITION_KEY]
            else:
                tddf = df.native
                keys = partition_spec.partition_by
            result = self.execution_engine.pl_utils.safe_groupby_apply(  # type: ignore
                tddf, keys, _map, meta=meta  # type: ignore
            )
        return DaskDataFrame(result, output_schema)


class DaskExecutionEngine(ExecutionEngine):
    """The execution engine based on `Dask <https://docs.dask.org/>`_.

    Please read |ExecutionEngineTutorial| to understand this important Fugue concept

    :param dask_client: Dask distributed client, defaults to None. If None, then it
      will try to get the current active global client. If there is no active client,
      it will create and use a global `Client(processes=True)`
    :param conf: |ParamsLikeObject| defaults to None, read |FugueConfig| to
      learn Fugue specific options

    .. note::

        You should setup Dask single machine or distributed environment in the
        :doc:`common <dask:setup>` way.
        Before initializing :class:`~.DaskExecutionEngine`
    """

    def __init__(self, dask_client: Optional[Client] = None, conf: Any = None):
        p = ParamDict(FUGUE_DASK_DEFAULT_CONF)
        p.update(ParamDict(conf))
        super().__init__(p)
        self._fs = FileSystem()
        self._log = logging.getLogger()
        self._client = DASK_UTILS.get_or_create_client(dask_client)
        self._native = NativeExecutionEngine(conf=conf)

    def __repr__(self) -> str:
        return "DaskExecutionEngine"

    @property
    def is_distributed(self) -> bool:
        return True

    @property
    def dask_client(self) -> Client:
        """The Dask Client associated with this engine"""
        return self._client

    @property
    def log(self) -> logging.Logger:
        return self._log

    @property
    def fs(self) -> FileSystem:
        return self._fs

    def create_default_sql_engine(self) -> SQLEngine:
        return QPDDaskEngine(self)

    def create_default_map_engine(self) -> MapEngine:
        return DaskMapEngine(self)

    def get_current_parallelism(self) -> int:
        res = dict(self.dask_client.nthreads())
        return sum(res.values())

    @property
    def pl_utils(self) -> DaskUtils:
        """Pandas-like dataframe utils"""
        return DaskUtils()

    def to_df(self, df: Any, schema: Any = None) -> DaskDataFrame:
        """Convert a data structure to :class:`~.fugue_dask.dataframe.DaskDataFrame`

        :param data: :class:`~.fugue.dataframe.dataframe.DataFrame`,
          :class:`dask:dask.dataframe.DataFrame`,
          pandas DataFrame or list or iterable of arrays
        :param schema: |SchemaLikeObject|, defaults to None.
        :return: engine compatible dataframe

        .. note::

            * if the input is already :class:`~.fugue_dask.dataframe.DaskDataFrame`,
              it should return itself
            * For list or iterable of arrays, ``schema`` must be specified
            * When ``schema`` is not None, a potential type cast may happen to ensure
              the dataframe's schema.
            * all other methods in the engine can take arbitrary dataframes and
              call this method to convert before doing anything
        """

        return to_dask_engine_df(df, schema)

    def repartition(
        self, df: DataFrame, partition_spec: PartitionSpec
    ) -> DaskDataFrame:
        df = self.to_df(df)
        if partition_spec.empty:
            return df
        if len(partition_spec.partition_by) > 0:
            return df
        p = partition_spec.get_num_partitions(
            **{
                KEYWORD_ROWCOUNT: lambda: df.persist().count(),  # type: ignore
                KEYWORD_PARALLELISM: lambda: self.get_current_parallelism(),
            }
        )
        if p > 0:
            if partition_spec.algo == "even":
                pdf = df.as_pandas()
                ddf = dd.from_pandas(pdf, npartitions=p, sort=False)
            else:
                ddf = df.native.repartition(npartitions=p)
            return DaskDataFrame(ddf, schema=df.schema, type_safe=False)
        return df

    def broadcast(self, df: DataFrame) -> DataFrame:
        return self.to_df(df)

    def persist(
        self,
        df: DataFrame,
        lazy: bool = False,
        **kwargs: Any,
    ) -> DataFrame:
        res = self.to_df(df)
        res.reset_metadata(df.metadata)
        return res.persist()

    def join(
        self,
        df1: DataFrame,
        df2: DataFrame,
        how: str,
        on: Optional[List[str]] = None,
    ) -> DataFrame:
        key_schema, output_schema = get_join_schemas(df1, df2, how=how, on=on)
        d = self.pl_utils.join(
            self.to_df(df1).native,
            self.to_df(df2).native,
            join_type=how,
            on=key_schema.names,
        )
        return DaskDataFrame(d, output_schema)

    def union(
        self,
        df1: DataFrame,
        df2: DataFrame,
        distinct: bool = True,
    ) -> DataFrame:
        assert_or_throw(
            df1.schema == df2.schema,
            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
        )
        d = self.pl_utils.union(
            self.to_df(df1).native, self.to_df(df2).native, unique=distinct
        )
        return DaskDataFrame(d, df1.schema)

    def subtract(
        self,
        df1: DataFrame,
        df2: DataFrame,
        distinct: bool = True,
    ) -> DataFrame:
        assert_or_throw(
            distinct, NotImplementedError("EXCEPT ALL for DaskExecutionEngine")
        )
        assert_or_throw(
            df1.schema == df2.schema,
            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
        )
        d = self.pl_utils.except_df(
            self.to_df(df1).native, self.to_df(df2).native, unique=distinct
        )
        return DaskDataFrame(d, df1.schema)

    def intersect(
        self,
        df1: DataFrame,
        df2: DataFrame,
        distinct: bool = True,
    ) -> DataFrame:
        assert_or_throw(
            distinct, NotImplementedError("INTERSECT ALL for DaskExecutionEngine")
        )
        assert_or_throw(
            df1.schema == df2.schema,
            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
        )
        d = self.pl_utils.intersect(
            self.to_df(df1).native, self.to_df(df2).native, unique=distinct
        )
        return DaskDataFrame(d, df1.schema)

    def distinct(self, df: DataFrame) -> DataFrame:
        d = self.pl_utils.drop_duplicates(self.to_df(df).native)
        return DaskDataFrame(d, df.schema)

    def dropna(
        self,
        df: DataFrame,
        how: str = "any",
        thresh: int = None,
        subset: List[str] = None,
    ) -> DataFrame:
        kw: Dict[str, Any] = dict(how=how)
        if thresh is not None:
            kw["thresh"] = thresh
        if subset is not None:
            kw["subset"] = subset
        if how == "any" and thresh is not None:
            del kw["how"]  # to deal with a dask logic flaw
        d = self.to_df(df).native.dropna(**kw)
        return DaskDataFrame(d, df.schema)

    def fillna(self, df: DataFrame, value: Any, subset: List[str] = None) -> DataFrame:
        assert_or_throw(
            (not isinstance(value, list)) and (value is not None),
            ValueError("fillna value can not be a list or None"),
        )
        if isinstance(value, dict):
            assert_or_throw(
                (None not in value.values()) and (any(value.values())),
                ValueError(
                    "fillna dict can not contain None and needs at least one value"
                ),
            )
            mapping = value
        else:
            # If subset is none, apply to all columns
            subset = subset or df.columns
            mapping = {col: value for col in subset}
        d = self.to_df(df).native.fillna(mapping)
        return DaskDataFrame(d, df.schema)

    def sample(
        self,
        df: DataFrame,
        n: Optional[int] = None,
        frac: Optional[float] = None,
        replace: bool = False,
        seed: Optional[int] = None,
    ) -> DataFrame:
        assert_or_throw(
            (n is None and frac is not None) or (n is not None and frac is None),
            ValueError("one and only one of n and frac should be set"),
        )
        # TODO: dask does not support sample by number of rows
        d = self.to_df(df).native.sample(
            n=n, frac=frac, replace=replace, random_state=seed
        )
        return DaskDataFrame(d, df.schema)

    def take(
        self,
        df: DataFrame,
        n: int,
        presort: str,
        na_position: str = "last",
        partition_spec: Optional[PartitionSpec] = None,
    ) -> DataFrame:
        partition_spec = partition_spec or PartitionSpec()
        assert_or_throw(
            isinstance(n, int),
            ValueError("n needs to be an integer"),
        )
        d = self.to_df(df).native
        meta = [(d[x].name, d[x].dtype) for x in d.columns]

        if presort:
            presort = parse_presort_exp(presort)
        # Use presort over partition_spec.presort if possible
        _presort: IndexedOrderedDict = presort or partition_spec.presort

        def _partition_take(partition, n, presort):
            if len(presort.keys()) > 0:
                partition = partition.sort_values(
                    list(presort.keys()),
                    ascending=list(presort.values()),
                    na_position=na_position,
                )
            return partition.head(n)

        if len(partition_spec.partition_by) == 0:
            if len(_presort.keys()) == 0:
                d = d.head(n)
            else:
                # Use the default partition
                d = (
                    d.map_partitions(_partition_take, n, _presort, meta=meta)
                    .reset_index(drop=True)
                    .compute()
                )
                # compute() brings this to Pandas so we can use pandas
                d = d.sort_values(
                    list(_presort.keys()),
                    ascending=list(_presort.values()),
                    na_position=na_position,
                ).head(n)

        else:
            d = (
                d.groupby(partition_spec.partition_by, dropna=False)
                .apply(_partition_take, n=n, presort=_presort, meta=meta)
                .reset_index(drop=True)
            )

        return DaskDataFrame(d, df.schema)

    def load_df(
        self,
        path: Union[str, List[str]],
        format_hint: Any = None,
        columns: Any = None,
        **kwargs: Any,
    ) -> DaskDataFrame:
        return self.to_df(
            load_df(
                path, format_hint=format_hint, columns=columns, fs=self.fs, **kwargs
            )
        )

    def save_df(
        self,
        df: DataFrame,
        path: str,
        format_hint: Any = None,
        mode: str = "overwrite",
        partition_spec: Optional[PartitionSpec] = None,
        force_single: bool = False,
        **kwargs: Any,
    ) -> None:
        partition_spec = partition_spec or PartitionSpec()
        if force_single:
            self._native.save_df(
                df,
                path=path,
                format_hint=format_hint,
                mode=mode,
                partition_spec=partition_spec,
                force_single=force_single,
                **kwargs,
            )
        else:
            if not partition_spec.empty:
                kwargs["partition_on"] = partition_spec.partition_by
            self.fs.makedirs(os.path.dirname(path), recreate=True)
            df = self.to_df(df)
            save_df(df, path, format_hint=format_hint, mode=mode, fs=self.fs, **kwargs)


def to_dask_engine_df(df: Any, schema: Any = None) -> DaskDataFrame:
    """Convert a data structure to :class:`~.fugue_dask.dataframe.DaskDataFrame`

    :param data: :class:`~.fugue.dataframe.dataframe.DataFrame`,
      :class:`dask:dask.dataframe.DataFrame`,
      pandas DataFrame or list or iterable of arrays
    :param schema: |SchemaLikeObject|, defaults to None.
    :return: engine compatible dataframe

    .. note::

        * if the input is already :class:`~fugue_dask.dataframe.DaskDataFrame`,
          it should return itself
        * For list or iterable of arrays, ``schema`` must be specified
        * When ``schema`` is not None, a potential type cast may happen to ensure
          the dataframe's schema.
        * all other methods in the engine can take arbitrary dataframes and
          call this method to convert before doing anything
    """

    if isinstance(df, DataFrame):
        assert_or_throw(
            schema is None,
            ValueError("schema must be None when df is a DataFrame"),
        )
        if isinstance(df, DaskDataFrame):
            return df
        if isinstance(df, PandasDataFrame):
            res = DaskDataFrame(df.native, df.schema)
        else:
            res = DaskDataFrame(df.as_array(type_safe=True), df.schema)
        res.reset_metadata(df.metadata)
        return res
    return DaskDataFrame(df, schema)



from typing import Any, Dict, List, Union, Callable

from fugue.dataframe.dataframe import DataFrame
from triad.collections.dict import IndexedOrderedDict
from triad.exceptions import InvalidOperationError
from triad.utils.assertion import assert_or_throw


class DataFrames(IndexedOrderedDict[str, DataFrame]):
    """Ordered dictionary of DataFrames. There are two modes: with keys
    and without keys. If without key ``_<n>`` will be used as the key
    for each dataframe, and it will be treated as an array in Fugue framework.

    It's a subclass of dict, so it supports all dict operations. It's
    also ordered, so you can trust the order of keys and values.

    The initialization is flexible

    >>> df1 = ArrayDataFrame([[0]],"a:int")
    >>> df2 = ArrayDataFrame([[1]],"a:int")
    >>> dfs = DataFrames(df1,df2)  # init as [df1, df2]
    >>> assert not dfs.has_key
    >>> assert df1 is dfs[0] and df2 is dfs[1]
    >>> dfs_array = list(dfs.values())
    >>> dfs = DataFrames(a=df1,b=df2)  # init as {a:df1, b:df2}
    >>> assert dfs.has_key
    >>> assert df1 is dfs[0] and df2 is dfs[1]  # order is guaranteed
    >>> df3 = ArrayDataFrame([[1]],"b:int")
    >>> dfs2 = DataFrames(dfs, c=df3)  # {a:df1, b:df2, c:df3}
    >>> dfs2 = DataFrames(dfs, df3)  # invalid, because dfs has key, df3 doesn't
    >>> dfs2 = DataFrames(dict(a=df1,b=df2))  # init as {a:df1, b:df2}
    >>> dfs2 = DataFrames([df1,df2],df3)  # init as [df1, df2, df3]
    """

    def __init__(self, *args: Any, **kwargs: Any):  # noqa: C901
        super().__init__()
        self._has_key: bool = False
        for d in args:
            if isinstance(d, DataFrames):
                if d.has_key:
                    for k, v in d.items():
                        self[k] = v
                else:
                    for v in d.values():
                        self._append(v)
            elif isinstance(d, DataFrame):
                self._append(d)
            elif isinstance(d, List):
                for o in d:
                    if isinstance(o, tuple):
                        self[o[0]] = o[1]
                    else:
                        self._append(o)
            elif isinstance(d, Dict):
                self.update(d)
            else:
                raise ValueError(f"{d} is not valid to initialize DataFrames")
        self.update(kwargs)
        self.set_readonly()

    @property
    def has_key(self):
        """If this collection has key (dict-like) or not (list-like)"""
        return self._has_key

    def __setitem__(  # type: ignore
        self, key: str, value: DataFrame, *args: Any, **kwds: Any
    ) -> None:
        assert isinstance(key, str)
        assert_or_throw(
            len(self) == 0 or self.has_key,
            InvalidOperationError("this DataFrames can's have key"),
        )
        assert_or_throw(
            isinstance(value, DataFrame),
            lambda: ValueError(f"{key} has non DataFrame value"),
        )
        super().__setitem__(key, value, *args, **kwds)  # type: ignore
        self._has_key = True

    def __getitem__(self, key: Union[str, int]) -> DataFrame:  # type: ignore
        if isinstance(key, int):
            key = self.get_key_by_index(key)
        return super().__getitem__(key)  # type: ignore

    def convert(self, func: Callable[["DataFrame"], DataFrame]) -> "DataFrames":
        """Create another DataFrames with the same structure,
        but all converted by ``func``

        :return: the new DataFrames

        .. admonition:: Examples

            >>> dfs2 = dfs.convert(lambda df: df.as_local()) # convert all to local
        """
        if self.has_key:
            return DataFrames([(k, func(v)) for k, v in self.items()])
        else:
            return DataFrames([func(v) for v in self.values()])

    def _append(self, value: Any):
        assert_or_throw(
            not self.has_key, InvalidOperationError("this DataFrames must have key")
        )
        assert_or_throw(
            isinstance(value, DataFrame),
            lambda: ValueError(f"{value} is not a DataFrame"),
        )
        super().__setitem__("_" + str(len(self)), value)



from typing import Any, Dict, Iterable, List, Optional

import pandas as pd
import pyarrow as pa
from triad import Schema, assert_or_throw
from triad.utils.iter import EmptyAwareIterable, make_empty_aware

from fugue.exceptions import FugueDataFrameInitError

from .array_dataframe import ArrayDataFrame
from .arrow_dataframe import ArrowDataFrame
from .dataframe import (
    DataFrame,
    LocalBoundedDataFrame,
    LocalDataFrame,
    LocalUnboundedDataFrame,
)
from .pandas_dataframe import PandasDataFrame


class LocalDataFrameIterableDataFrame(LocalUnboundedDataFrame):
    """DataFrame that wraps an iterable of local dataframes

    :param df: an iterable of
      :class:`~fugue.dataframe.dataframe.DataFrame`. If any is not local,
      they will be converted to :class:`~fugue.dataframe.dataframe.LocalDataFrame`
      by :meth:`~fugue.dataframe.dataframe.DataFrame.as_local`
    :param schema: |SchemaLikeObject|, if it is provided, it must match the schema
      of the dataframes

    .. admonition:: Examples

        .. code-block:: python

            def get_dfs(seq):
                yield IterableDataFrame([], "a:int,b:int")
                yield IterableDataFrame([[1, 10]], "a:int,b:int")
                yield ArrayDataFrame([], "a:int,b:str")

            df = LocalDataFrameIterableDataFrame(get_dfs())
            for subdf in df.native:
                subdf.show()

    .. note::

        It's ok to peek the dataframe, it will not affect the iteration, but it's
        invalid to count.

        ``schema`` can be used when the iterable contains no dataframe. But if there
        is any dataframe, ``schema`` must match the schema of the dataframes.

        For the iterable of dataframes, if there is any empty dataframe, they will
        be skipped and their schema will not matter. However, if all dataframes
        in the interable are empty, then the last empty dataframe will be used to
        set the schema.
    """

    def __init__(self, df: Any = None, schema: Any = None):  # noqa: C901
        if isinstance(df, Iterable):
            self._native = make_empty_aware(self._dfs_wrapper(df))
            orig_schema: Optional[Schema] = None
            if not self._native.empty:
                orig_schema = self._native.peek().schema
        else:
            raise ValueError(
                f"{df} is incompatible with LocalDataFrameIterableDataFrame"
            )
        if orig_schema is None and schema is None:
            raise FugueDataFrameInitError(
                "schema is not provided and the input is empty"
            )
        elif orig_schema is None and schema is not None:
            pass
        elif orig_schema is not None and schema is None:
            schema = orig_schema
        else:
            schema = Schema(schema) if not isinstance(schema, Schema) else schema
            assert_or_throw(
                orig_schema == schema,
                lambda: f"iterable schema {orig_schema} is different from {schema}",
            )
        super().__init__(schema)

    def _dfs_wrapper(self, dfs: Iterable[DataFrame]) -> Iterable[LocalDataFrame]:
        last_empty: Any = None
        last_schema: Any = None
        yielded = False
        for df in dfs:
            if df.empty:
                last_empty = df
            else:
                assert_or_throw(
                    last_schema is None or df.schema == last_schema,
                    lambda: FugueDataFrameInitError(
                        f"encountered schema {df.schema} doesn't match"
                        f" the original schema {df.schema}"
                    ),
                )
                if last_schema is None:
                    last_schema = df.schema
                yield df.as_local()
                yielded = True
        if not yielded and last_empty is not None:
            yield last_empty

    @property
    def native(self) -> EmptyAwareIterable[LocalDataFrame]:
        """Iterable of dataframes"""
        return self._native

    @property
    def empty(self) -> bool:
        return self.native.empty or self.native.peek().empty

    def peek_array(self) -> List[Any]:
        self.assert_not_empty()
        return self.native.peek().peek_array()

    def _select_cols(self, keys: List[Any]) -> DataFrame:
        if self.empty:
            return ArrayDataFrame([], self.schema)[keys]

        def _transform():
            for df in self.native:
                yield df[keys]

        return LocalDataFrameIterableDataFrame(_transform())

    def rename(self, columns: Dict[str, str]) -> DataFrame:
        if self.empty:
            return ArrayDataFrame([], self.schema).rename(columns)

        def _transform() -> Iterable[DataFrame]:
            for df in self.native:
                yield df.rename(columns)

        return LocalDataFrameIterableDataFrame(_transform())

    def alter_columns(self, columns: Any) -> DataFrame:
        if self.empty:
            return ArrayDataFrame([], self.schema).alter_columns(columns)

        def _transform() -> Iterable[DataFrame]:
            for df in self.native:
                yield df.alter_columns(columns)

        return LocalDataFrameIterableDataFrame(_transform())

    def as_local_bounded(self) -> "LocalBoundedDataFrame":
        return ArrowDataFrame(self.as_arrow())

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        return sum(
            (df.as_array(columns=columns, type_safe=type_safe) for df in self.native),
            [],
        )

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        for df in self.native:
            yield from df.as_array_iterable(columns=columns, type_safe=type_safe)

    def as_pandas(self) -> pd.DataFrame:
        if self.empty:
            return ArrayDataFrame([], self.schema).as_pandas()

        return pd.concat(df.as_pandas() for df in self.native)

    def as_arrow(self, type_safe: bool = False) -> pa.Table:
        if self.empty:
            return ArrayDataFrame([], self.schema).as_arrow()

        return pa.concat_tables(df.as_arrow() for df in self.native)

    def head(
        self, n: int, columns: Optional[List[str]] = None
    ) -> LocalBoundedDataFrame:
        res: List[Any] = []
        for row in self.as_array_iterable(columns, type_safe=True):
            if n < 1:
                break
            res.append(list(row))
            n -= 1
        return ArrayDataFrame(
            res, self.schema if columns is None else self.schema.extract(columns)
        )

    def _drop_cols(self, cols: List[str]) -> DataFrame:
        if self.empty:
            return ArrayDataFrame([], self.schema)._drop_cols(cols)

        def _transform() -> Iterable[DataFrame]:
            for df in self.native:
                yield df._drop_cols(cols)

        return LocalDataFrameIterableDataFrame(_transform())


class IterablePandasDataFrame(LocalDataFrameIterableDataFrame):
    def as_local_bounded(self) -> "LocalBoundedDataFrame":
        return PandasDataFrame(self.as_pandas(), schema=self.schema)


class IterableArrowDataFrame(LocalDataFrameIterableDataFrame):
    pass



import json
from datetime import datetime
from typing import Any

import numpy as np
import pandas as pd
from fugue.dataframe import IterableDataFrame, PandasDataFrame
from fugue.dataframe.utils import _df_eq as df_eq
from fugue_test.dataframe_suite import DataFrameTests
from pytest import raises
from triad.collections.schema import Schema, SchemaError
from triad.exceptions import InvalidOperationError


class IterableDataFrameTests(DataFrameTests.Tests):
    def df(
        self, data: Any = None, schema: Any = None
    ) -> IterableDataFrame:
        return IterableDataFrame(data, schema)


def test_init():
    df = IterableDataFrame(schema="a:str,b:int")
    assert df.empty
    assert df.schema == "a:str,b:int"
    assert not df.is_bounded

    data = [["a", 1], ["b", 2]]
    df = IterableDataFrame(data, "a:str,b:str")
    assert [["a", "1"], ["b", "2"]] == df.as_array(type_safe=True)
    assert df.empty  # after iterating all items
    df = IterableDataFrame(data, "a:str,b:int")
    assert [["a", 1], ["b", 2]] == df.as_array(type_safe=True)
    df = IterableDataFrame(data, "a:str,b:double")
    assert [["a", 1.0], ["b", 2.0]] == df.as_array(type_safe=True)

    df = IterableDataFrame(data, "a:str,b:double")
    ddf = IterableDataFrame(df)
    assert [["a", 1.0], ["b", 2.0]] == ddf.as_array(type_safe=True)
    df = IterableDataFrame(data, "a:str,b:double")
    ddf = IterableDataFrame(df, "a:str,b:float64")
    assert [["a", 1.0], ["b", 2.0]] == ddf.as_array(type_safe=True)
    df = IterableDataFrame(data, "a:str,b:double")
    ddf = IterableDataFrame(df, "b:str,a:str")
    assert [["1", "a"], ["2", "b"]] == ddf.as_array(type_safe=True)
    df = IterableDataFrame(data, "a:str,b:double")
    ddf = IterableDataFrame(df, ["b"])
    assert ddf.schema == "b:double"
    assert [[1.0], [2.0]] == ddf.as_array(type_safe=True)
    df = IterableDataFrame(data, "a:str,b:double")
    ddf = IterableDataFrame(df, ["a:str,b:str"])
    assert [["a", "1"], ["b", "2"]] == ddf.as_array(type_safe=True)
    df = IterableDataFrame(data, "a:str,b:double")
    ddf = IterableDataFrame(df, ["b:str"])
    assert [["1"], ["2"]] == ddf.as_array(type_safe=True)

    pdf = PandasDataFrame(data, "a:str,b:double")
    df = IterableDataFrame(pdf, "a:str,b:double")
    assert [["a", 1.0], ["b", 2.0]] == df.as_array(type_safe=True)
    df = IterableDataFrame(pdf, "b:str,a:str")
    assert [["1.0", "a"], ["2.0", "b"]] == df.as_array(type_safe=True)

    df = IterableDataFrame([], "x:str,y:double")
    assert df.empty
    assert df.is_local

    raises(Exception, lambda: IterableDataFrame(123))


def test_simple_methods():
    df = IterableDataFrame([["a", 1], ["b", "2"]], "x:str,y:double")
    raises(InvalidOperationError, lambda: df.count())
    assert not df.empty
    assert ["a", 1.0] == df.peek_array()
    assert dict(x="a", y=1.0) == df.peek_dict()
    assert [["a", 1], ["b", "2"]] == df.as_array()

    df = IterableDataFrame([["a", 1], ["b", "2"]], "x:str,y:double")
    pdf = df.as_pandas()
    assert [["a", 1.0], ["b", 2.0]] == pdf.values.tolist()

    df = IterableDataFrame([], "x:str,y:double")
    pdf = df.as_pandas()
    assert [] == pdf.values.tolist()


def test_nested():
    data = [[dict(a=1, b=[3, 4], d=1.0)], [json.dumps(dict(b=[30, "40"]))]]
    df = IterableDataFrame(data, "a:{a:str,b:[int]}")
    a = df.as_array(type_safe=True)
    assert [[dict(a="1", b=[3, 4])], [dict(a=None, b=[30, 40])]] == a

    data = [[[json.dumps(dict(b=[30, "40"]))]]]
    df = IterableDataFrame(data, "a:[{a:str,b:[int]}]")
    a = df.as_array(type_safe=True)
    assert [[[dict(a=None, b=[30, 40])]]] == a


def _test_as_array_perf():
    s = Schema()
    arr = []
    for i in range(100):
        s.append(f"a{i}:int")
        arr.append(i)
    for i in range(100):
        s.append(f"b{i}:int")
        arr.append(float(i))
    for i in range(100):
        s.append(f"c{i}:str")
        arr.append(str(i))
    data = []
    for i in range(5000):
        data.append(list(arr))
    df = IterableDataFrame(data, s)
    res = df.as_array()
    res = df.as_array(type_safe=True)
    nts, ts = 0.0, 0.0
    for i in range(10):
        t = datetime.now()
        res = df.as_array()
        nts += (datetime.now() - t).total_seconds()
        t = datetime.now()
        res = df.as_array(type_safe=True)
        ts += (datetime.now() - t).total_seconds()
    print(nts, ts)



import pandas as pd
import ray
import ray.data as rd
from fugue import FugueWorkflow

from fugue_ray import RayExecutionEngine


def test_registry():
    def creator() -> rd.Dataset:
        return rd.from_pandas(pd.DataFrame(dict(a=[1, 2], b=["a", "b"])))

    def processor1(ctx: RayExecutionEngine, df: rd.Dataset) -> pd.DataFrame:
        assert isinstance(ctx, RayExecutionEngine)
        return df.to_pandas()

    def processor2(df: pd.DataFrame) -> rd.Dataset:
        return rd.from_pandas(df)

    def outputter(df: rd.Dataset) -> None:
        assert [[1, "a"], [2, "b"]] == df.to_pandas().values.tolist()

    dag = FugueWorkflow()
    dag.create(creator).process(processor1).process(processor2).output(outputter)

    with ray.init():
        dag.run("ray")



from typing import Any, List

import numpy as np
import pandas as pd
import pyspark
import pyspark.rdd as pr
import pyspark.sql as ps
import pytest
from pyspark import SparkContext, StorageLevel
from pyspark.sql import DataFrame as SDataFrame
from pyspark.sql import SparkSession
from pytest import raises
from triad import Schema

import fugue.api as fa
from fugue import transform
from fugue.collections.partition import PartitionSpec
from fugue.dataframe import (
    ArrayDataFrame,
    ArrowDataFrame,
    IterablePandasDataFrame,
    LocalDataFrameIterableDataFrame,
    PandasDataFrame,
)
from fugue.dataframe.utils import _df_eq as df_eq
from fugue.extensions.transformer import Transformer, transformer
from fugue.plugins import infer_execution_engine
from fugue.workflow.workflow import FugueWorkflow
from fugue_spark.dataframe import SparkDataFrame
from fugue_spark.execution_engine import SparkExecutionEngine
from fugue_test.builtin_suite import BuiltInTests
from fugue_test.execution_suite import ExecutionEngineTests


class SparkExecutionEngineTests(ExecutionEngineTests.Tests):
    @pytest.fixture(autouse=True)
    def init_session(self, spark_session):
        self.spark_session = spark_session

    def make_engine(self):
        session = SparkSession.builder.getOrCreate()
        e = SparkExecutionEngine(
            session, {"test": True, "fugue.spark.use_pandas_udf": False}
        )
        return e

    def test_properties(self):
        assert self.engine.is_distributed
        assert self.engine.map_engine.is_distributed
        assert self.engine.sql_engine.is_distributed

    def test_get_parallelism(self):
        assert fa.get_current_parallelism() == 4

    def test_not_using_pandas_udf(self):
        assert not self.engine.create_default_map_engine()._should_use_pandas_udf(
            Schema("a:int")
        )

    def test__join_outer_pandas_incompatible(self):
        return

    def test_to_df(self):
        e = self.engine
        o = ArrayDataFrame([[1, 2], [None, 3]], "a:double,b:int")
        a = e.to_df(o)
        assert a is not o
        res = a.native.collect()
        assert res[0][0] == 1.0 or res[0][0] is None
        assert res[1][0] == 1.0 or res[1][0] is None
        df_eq(a, o, throw=True)

        o = ArrowDataFrame([[1, 2], [None, 3]], "a:double,b:int")
        a = e.to_df(o)
        assert a is not o
        res = a.native.collect()
        assert res[0][0] == 1.0 or res[0][0] is None
        assert res[1][0] == 1.0 or res[1][0] is None

        a = e.to_df([[1, None]], "a:int,b:int")
        df_eq(a, [[1, None]], "a:int,b:int", throw=True)

        o = PandasDataFrame([[{"a": "b"}, 2]], "a:{a:str},b:int")
        a = e.to_df(o)
        assert a is not o
        res = a.as_array(type_safe=True)
        assert res[0][0] == {"a": "b"}

        pdf = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
        pdf = pdf[pdf.a < 1]
        a = e.to_df(pdf)
        assert fa.get_schema(a) == "a:long,b:long"

    def test_persist(self):
        e = self.engine

        o = ArrayDataFrame([[1, 2]], "a:int,b:int")
        o.reset_metadata({"a": 1})
        a = e.persist(o)
        assert a.metadata == {"a": 1}
        df_eq(a, o, throw=True)
        a = e.persist(o, level=StorageLevel.MEMORY_ONLY)
        df_eq(a, o, throw=True)
        a = e.persist(o, level="MEMORY_ONLY")
        df_eq(a, o, throw=True)
        # this passed because persist is run once on ths same object
        e.persist(a, level="xyz")
        raises(ValueError, lambda: e.persist(o, level="xyz"))

    def test_sample_n(self):
        engine = self.engine
        a = engine.to_df([[x] for x in range(100)], "a:int")

        with raises(NotImplementedError):
            # replace is not allowed
            engine.sample(a, n=90, replace=True)

        b = engine.sample(a, n=90)
        assert abs(len(b.as_array()) - 90) < 2

    def test_infer_engine(self):
        df = self.spark_session.createDataFrame(pd.DataFrame([[0]], columns=["a"]))
        assert isinstance(infer_execution_engine([df]), SparkSession)

        fdf = SparkDataFrame(df)
        assert isinstance(infer_execution_engine([fdf]), SparkSession)


@pytest.mark.skipif(pyspark.__version__ < "3", reason="pyspark < 3")
class SparkExecutionEnginePandasUDFTests(ExecutionEngineTests.Tests):
    @pytest.fixture(autouse=True)
    def init_session(self, spark_session):
        self.spark_session = spark_session

    def make_engine(self):
        session = SparkSession.builder.getOrCreate()
        e = SparkExecutionEngine(session, {"test": True})
        return e

    def test_get_parallelism(self):
        assert fa.get_current_parallelism() == 4

    def test__join_outer_pandas_incompatible(self):
        return

    def test_using_pandas_udf(self):
        assert self.engine.map_engine._should_use_pandas_udf(  # type: ignore
            Schema("a:int")
        )
        assert not self.engine.map_engine._should_use_pandas_udf(  # type: ignore
            Schema("a:{x:int}")
        )

    def test_sample_n(self):
        engine = self.engine
        a = engine.to_df([[x] for x in range(100)], "a:int")

        b = engine.sample(a, n=90)
        assert abs(len(b.as_array()) - 90) < 2

    def test_map_in_pandas(self):
        if not hasattr(ps.DataFrame, "mapInPandas"):
            return

        def add(cursor, data):
            # assertion for pandas udf input
            assert isinstance(data, LocalDataFrameIterableDataFrame)

            def get_dfs():
                for df in data.native:
                    pdf = df.as_pandas()
                    pdf["zz"] = pdf["xx"] + pdf["yy"]
                    yield PandasDataFrame(pdf)

            return IterablePandasDataFrame(get_dfs())

        e = self.engine
        np.random.seed(0)
        df = pd.DataFrame(np.random.randint(0, 5, (100000, 2)), columns=["xx", "yy"])
        expected = PandasDataFrame(df.assign(zz=df.xx + df.yy), "xx:int,yy:int,zz:int")
        a = e.to_df(df)
        # no partition
        c = e.map_engine.map_dataframe(
            a, add, "xx:int,yy:int,zz:int", PartitionSpec(num=16)
        )
        df_eq(c, expected, throw=True)


class SparkExecutionEngineBuiltInTests(BuiltInTests.Tests):
    @pytest.fixture(autouse=True)
    def init_session(self, spark_session):
        self.spark_session = spark_session

    def make_engine(self):
        session = SparkSession.builder.getOrCreate()
        e = SparkExecutionEngine(
            session,
            {
                "test": True,
                "fugue.spark.use_pandas_udf": False,
                "fugue.rpc.server": "fugue.rpc.flask.FlaskRPCServer",
                "fugue.rpc.flask_server.host": "127.0.0.1",
                "fugue.rpc.flask_server.port": "1234",
                "fugue.rpc.flask_server.timeout": "2 sec",
                "spark.sql.shuffle.partitions": "10",
            },
        )
        return e

    def test_df_init(self):
        sdf = self.spark_session.createDataFrame([[1.1]], "a:double")
        a = FugueWorkflow().df(sdf)
        df_eq(a.compute(SparkExecutionEngine), [[1.1]], "a:double")

    def test_yield_table(self):
        pass

    def test_default_session(self):
        a = FugueWorkflow().df([[0]], "a:int")
        df_eq(a.compute(SparkExecutionEngine), [[0]], "a:int")

    def test_repartition(self):
        with FugueWorkflow() as dag:
            a = dag.df(
                [[0, 1], [0, 2], [0, 3], [0, 4], [1, 1], [1, 2], [1, 3]], "a:int,b:int"
            )
            c = a.per_row().transform(count_partition)
            dag.output(c, using=assert_match, params=dict(values=[1, 1, 1, 1, 1, 1, 1]))
            c = a.partition(algo="even", num="ROWCOUNT/2").transform(count_partition)
            dag.output(c, using=assert_match, params=dict(values=[3, 2, 2]))
            c = a.per_partition_by("a").transform(count_partition)
            dag.output(c, using=assert_match, params=dict(values=[3, 4]))
            c = a.partition(algo="even", by=["a"]).transform(AssertMaxNTransform)
            dag.output(c, using=assert_match, params=dict(values=[3, 4]))
            c = a.partition(num=1).transform(count_partition)
            dag.output(c, using=assert_match, params=dict(values=[7]))
            c = a.partition(algo="rand", num=100).transform(count_partition).persist()
        dag.run(self.engine)

    def test_repartition_large(self):
        with FugueWorkflow() as dag:
            a = dag.df([[p, 0] for p in range(100)], "a:int,b:int")
            c = (
                a.partition(algo="even", by=["a"])
                .transform(AssertMaxNTransform)
                .persist()
            )
            c = (
                a.partition(algo="even", num="ROWCOUNT/2", by=["a"])
                .transform(AssertMaxNTransform, params=dict(n=2))
                .persist()
            )
            c = (
                a.partition(algo="even", num="ROWCOUNT")
                .transform(AssertMaxNTransform)
                .persist()
            )
            dag.output(c, using=assert_all_n, params=dict(n=1, l=100))
            c = a.partition(algo="even", num="ROWCOUNT/2").transform(
                AssertMaxNTransform
            )
            dag.output(c, using=assert_all_n, params=dict(n=2, l=50))
            c = a.partition(num=1).transform(count_partition)
            dag.output(c, using=assert_match, params=dict(values=[100]))
        dag.run(self.engine)

    def test_coarse_partition(self):
        def verify_coarse_partition(df: pd.DataFrame) -> List[List[Any]]:
            ct = df.a.nunique()
            s = df.a * 1000 + df.b
            ordered = ((s - s.shift(1)).dropna() >= 0).all(axis=None)
            return [[ct, ordered]]

        def assert_(df: pd.DataFrame, rc: int, n: int, check_ordered: bool) -> None:
            if rc > 0:
                assert len(df) == rc
            assert df.ct.sum() == n
            if check_ordered:
                assert (df.ordered == True).all()

        gps = 100
        partition_num = 6
        df = pd.DataFrame(dict(a=list(range(gps)) * 10, b=range(gps * 10))).sample(
            frac=1.0
        )
        with FugueWorkflow() as dag:
            a = dag.df(df)
            c = a.partition(
                algo="coarse", by="a", presort="b", num=partition_num
            ).transform(verify_coarse_partition, schema="ct:int,ordered:bool")
            dag.output(
                c,
                using=assert_,
                params=dict(rc=partition_num, n=gps, check_ordered=True),
            )
        dag.run(self.engine)

    def test_session_as_engine(self):
        dag = FugueWorkflow()
        a = dag.df([[p, 0] for p in range(100)], "a:int,b:int")
        a.partition(algo="even", by=["a"]).transform(AssertMaxNTransform).persist()
        dag.run(self.spark_session)

    def test_interfaceless(self):
        sdf = self.spark_session.createDataFrame(
            [[1, 10], [0, 0], [1, 1], [0, 20]], "a int,b int"
        )

        # schema:*
        def f1(df: pd.DataFrame) -> pd.DataFrame:
            return df.sort_values("b").head(1)

        result = transform(sdf, f1, partition=dict(by=["a"]), engine=self.engine)
        assert isinstance(result, SDataFrame)
        assert result.toPandas().sort_values(["a"]).values.tolist() == [[0, 0], [1, 1]]

    def test_annotation_1(self):
        def m_c(engine: SparkExecutionEngine) -> ps.DataFrame:
            return engine.spark_session.createDataFrame([[0]], "a:long")

        def m_p(engine: SparkExecutionEngine, df: ps.DataFrame) -> ps.DataFrame:
            return df

        def m_o(engine: SparkExecutionEngine, df: ps.DataFrame) -> None:
            assert 1 == df.toPandas().shape[0]

        with FugueWorkflow() as dag:
            df = dag.create(m_c).process(m_p)
            df.assert_eq(dag.df([[0]], "a:long"))
            df.output(m_o)
        dag.run(self.engine)

    def test_annotation_2(self):
        def m_c(session: SparkSession) -> ps.DataFrame:
            return session.createDataFrame([[0]], "a:long")

        def m_p(session: SparkSession, df: ps.DataFrame) -> ps.DataFrame:
            assert isinstance(session, SparkSession)
            return df

        def m_o(session: SparkSession, df: ps.DataFrame) -> None:
            assert isinstance(session, SparkSession)
            assert 1 == df.toPandas().shape[0]

        with FugueWorkflow() as dag:
            df = dag.create(m_c).process(m_p)
            df.assert_eq(dag.df([[0]], "a:long"))
            df.output(m_o)
        dag.run(self.engine)

    def test_annotation_3(self):
        # schema: a:int,b:str
        def m_c(ctx: SparkContext) -> pr.RDD:
            return ctx.parallelize([(0, "x")])

        # schema: a:int,b:str
        def m_p(ctx: SparkContext, data: pr.RDD) -> pr.RDD:
            assert isinstance(ctx, SparkContext)
            return data

        def m_o(ctx: SparkContext, data: pr.RDD) -> None:
            assert isinstance(ctx, SparkContext)
            assert 1 == len(data.collect())

        with FugueWorkflow() as dag:
            df = dag.create(m_c).process(m_p)
            df.assert_eq(dag.df([[0, "x"]], "a:int,b:str"))
            df.output(m_o)
        dag.run(self.engine)


class SparkExecutionEnginePandasUDFBuiltInTests(SparkExecutionEngineBuiltInTests):
    @pytest.fixture(autouse=True)
    def init_session(self, spark_session):
        self.spark_session = spark_session

    def make_engine(self):
        session = SparkSession.builder.getOrCreate()
        e = SparkExecutionEngine(
            session,
            {
                "test": True,
                "fugue.spark.use_pandas_udf": True,
                "fugue.rpc.server": "fugue.rpc.flask.FlaskRPCServer",
                "fugue.rpc.flask_server.host": "127.0.0.1",
                "fugue.rpc.flask_server.port": "1234",
                "fugue.rpc.flask_server.timeout": "2 sec",
                "spark.sql.shuffle.partitions": "10",
            },
        )
        assert e.conf.get_or_throw("fugue.spark.use_pandas_udf", bool)
        return e


@transformer("ct:long")
def count_partition(df: List[List[Any]]) -> List[List[Any]]:
    return [[len(df)]]


class AssertMaxNTransform(Transformer):
    def get_output_schema(self, df):
        return "c:int"

    def transform(self, df):
        if not hasattr(self, "called"):
            self.called = 1
        else:
            self.called += 1
        n = self.params.get("n", 1)
        assert self.called <= n
        return ArrayDataFrame([[len(df.as_array())]], "c:int")


def assert_match(df: List[List[Any]], values: List[int]) -> None:
    assert set(values) == set(x[0] for x in df)


def assert_all_n(df: List[List[Any]], n, l) -> None:
    assert all(x[0] == n for x in df)
    assert l == len(df)



from typing import Any, Dict, Iterable, List

from fugue import FileSystem, FugueWorkflow, Schema
from fugue.execution.native_execution_engine import NativeExecutionEngine
from triad import to_uuid


def test_create():
    # by default, input data does not affect determinism
    id0 = FugueWorkflow().df([[0]], "a:int32").workflow.spec_uuid()
    id1 = FugueWorkflow().df([[1]], "a:int32").workflow.spec_uuid()
    id2 = FugueWorkflow().df([[1]], "a:int32").workflow.spec_uuid()

    assert id1 == id0
    assert id1 == id2

    # unless we use data_determiner
    id0 = (
        FugueWorkflow()
        .df([[0]], "a:int32", data_determiner=to_uuid)
        .workflow.spec_uuid()
    )
    id1 = (
        FugueWorkflow()
        .df([[1]], "a:int32", data_determiner=to_uuid)
        .workflow.spec_uuid()
    )
    id2 = (
        FugueWorkflow()
        .df([[1]], "a:int32", data_determiner=to_uuid)
        .workflow.spec_uuid()
    )

    assert id1 != id0
    assert id1 == id2


def test_checkpoint():
    id0 = FugueWorkflow().df([[0]], "a:int32").workflow.spec_uuid()
    id1 = FugueWorkflow().df([[0]], "a:int32").checkpoint().workflow.spec_uuid()
    # checkpoint doesn't change determinism
    assert id0 == id1


def test_yield():
    dag = FugueWorkflow()
    dag.df([[0]], "a:int32").show()
    id0 = dag.spec_uuid()
    x = FugueWorkflow().df([[0]], "a:int32")
    x.yield_file_as("x")
    x.show()
    id1 = x.workflow.spec_uuid()
    x = FugueWorkflow().df([[0]], "a:int32")
    x.deterministic_checkpoint().yield_file_as("y")
    x.show()
    id2 = x.workflow.spec_uuid()
    x = FugueWorkflow().df([[0]], "a:int32")
    x.deterministic_checkpoint().yield_dataframe_as("z")
    x.show()
    id3 = x.workflow.spec_uuid()
    # yield doesn't change determinism
    assert id0 == id1
    assert id0 == id2
    assert id0 == id3


def test_auto_persist():
    dag1 = FugueWorkflow()
    df1 = dag1.df([[0]], "a:int")
    df1.show()
    df1.show()
    id1 = dag1.spec_uuid()

    dag2 = FugueWorkflow({"fugue.workflow.auto_persist": True})
    df1 = dag2.df([[0]], "a:int")
    df1.show()
    df1.show()
    id2 = dag2.spec_uuid()

    dag3 = FugueWorkflow()
    df1 = dag3.df([[0]], "a:int").weak_checkpoint(level=None)
    df1.show()
    df1.show()
    id3 = dag3.spec_uuid()

    assert id1 == id2
    assert id2 == id3

    dag2 = FugueWorkflow(
        {
            "fugue.workflow.auto_persist": True,
            "fugue.workflow.auto_persist_value": "abc",
        }
    )
    df1 = dag2.df([[0]], "a:int")
    df1.show()
    df1.show()
    id2 = dag2.spec_uuid()

    dag3 = FugueWorkflow()
    df1 = dag3.df([[0]], "a:int").weak_checkpoint(level="abc")
    df1.show()
    df1.show()
    id3 = dag3.spec_uuid()

    assert id2 == id3

    dag1 = FugueWorkflow()
    df1 = dag1.df([[0]], "a:int")
    df1.show()
    id1 = dag1.spec_uuid()

    dag2 = FugueWorkflow({"fugue.workflow.auto_persist": True})
    df1 = dag2.df([[0]], "a:int")
    df1.show()  # auto persist will not trigger
    id2 = dag2.spec_uuid()

    dag3 = FugueWorkflow()
    df1 = dag3.df([[0]], "a:int").weak_checkpoint(level=None)
    df1.show()
    id3 = dag3.spec_uuid()

    assert id1 == id2
    assert id2 == id3  # checkpoint, including auto_persist doesn't change determinism


def test_workflow_determinism_1():
    dag1 = FugueWorkflow()
    a1 = dag1.create_data([[0], [0], [1]], "a:int32")
    b1 = a1.transform(mock_tf1, "*,b:int", pre_partition=dict(by=["a"], num=2))
    a1.show()

    dag2 = FugueWorkflow()
    a2 = dag2.create_data([[0], [0], [1]], "a:int32")
    b2 = a2.transform(mock_tf1, "*,b:int", pre_partition=dict(by=["a"], num=2))
    a2.show()

    assert dag1.spec_uuid() == dag1.spec_uuid()
    assert a1.spec_uuid() == a2.spec_uuid()
    assert b1.spec_uuid() == b2.spec_uuid()
    assert dag1.spec_uuid() == dag2.spec_uuid()


def test_workflow_determinism_2():
    dag1 = FugueWorkflow()
    dag1.create_data([[0], [0], [1]], "a:int32")  # <----
    a1 = dag1.create_data([[0], [0], [1]], "a:int32")
    b1 = a1.transform(mock_tf1, "*,b:int", pre_partition=dict(by=["a"], num=2))
    a1.show()

    dag2 = FugueWorkflow()
    a2 = dag2.create_data([[0], [0], [1]], "a:int32")
    b2 = a2.transform(mock_tf1, "*,b:int", pre_partition=dict(by=["a"], num=2))
    a2.show()

    assert a1.spec_uuid() == a2.spec_uuid()
    assert b1.spec_uuid() == b2.spec_uuid()
    assert dag1.spec_uuid() != dag2.spec_uuid()


def test_workflow_determinism_3():
    dag1 = FugueWorkflow()
    data = [[0], [0], [1]]
    a1 = dag1.create_data(data, "a:int32", data_determiner=to_uuid)
    b1 = a1.transform(mock_tf1, "*,b:int", pre_partition=dict(by=["a"], num=2))
    a1.show()

    dag2 = FugueWorkflow()
    data = [[1], [10], [20]]
    a2 = dag2.create_data(data, "a:int32", data_determiner=to_uuid)  # <---
    b2 = a2.transform(mock_tf1, "*,b:int", pre_partition=dict(by=["a"], num=2))
    a2.show()

    assert a1.spec_uuid() != a2.spec_uuid()
    assert b1.spec_uuid() != b2.spec_uuid()
    assert dag1.spec_uuid() != dag2.spec_uuid()


def test_workflow_determinism_4():
    dag1 = FugueWorkflow()
    a1 = dag1.create_data([[0], [0], [1]], "a:int32")
    b1 = a1.transform(mock_tf1, "*,b:int", pre_partition=dict(by=["a"], num=2))
    b1.out_transform(mock_tf1)
    a1.show()

    dag2 = FugueWorkflow()
    a2 = dag2.create_data([[0], [0], [1]], "a:int32")
    b2 = a2.transform(mock_tf1, "*,b:int", pre_partition=dict(by=["a"], num=20))  # <---
    b2.out_transform(mock_tf1)
    a2.show()

    assert a1.spec_uuid() == a2.spec_uuid()
    assert b1.spec_uuid() != b2.spec_uuid()
    assert dag1.spec_uuid() != dag2.spec_uuid()


def test_workflow_determinism_5():
    dag1 = FugueWorkflow()
    a1 = dag1.create_data([[0], [0], [1]], "a:int32")
    b1 = a1.transform(mock_tf1, "*,b:int", pre_partition=dict(by=["a"], num=2))
    a1.show()

    dag2 = FugueWorkflow()
    a2 = dag2.create_data([[0], [0], [1]], "a:int32")
    b2 = a2.transform(mock_tf1, "*,b:int", pre_partition=dict(by=["a"], num=2))
    a2.show(n=22)  # <---

    assert a1.spec_uuid() == a2.spec_uuid()
    assert b1.spec_uuid() == b2.spec_uuid()
    assert dag1.spec_uuid() != dag2.spec_uuid()


def test_workflow_determinism_6():
    dag1 = FugueWorkflow()
    dag1.create_data([[0], [0], [1]], "a:int32")
    a1 = dag1.create_data([[0], [0], [1]], "a:int32")
    b1 = dag1.create_data([[1], [0], [1]], "a:int32")
    c1 = a1.union(b1)

    dag2 = FugueWorkflow()
    a2 = dag1.create_data([[0], [0], [1]], "a:int32")
    b2 = dag1.create_data([[1], [0], [1]], "a:int32")
    c2 = a1.union(b2)

    assert a1.spec_uuid() == a2.spec_uuid()
    assert b1.spec_uuid() == b2.spec_uuid()
    assert c1.spec_uuid() == c2.spec_uuid()
    assert dag1.spec_uuid() != dag2.spec_uuid()


def test_workflow_determinism_7():
    dag1 = FugueWorkflow()
    a1 = dag1.create_data([[0], [0], [1]], "a:int32")
    a1.out_transform(mock_tf1)
    a1.show()

    dag2 = FugueWorkflow()
    a2 = dag2.create_data([[0], [0], [1]], "a:int32")
    a2.out_transform(mock_tf1)
    a2.show()

    dag3 = FugueWorkflow()
    a3 = dag3.create_data([[0], [0], [1]], "a:int32")
    a3.show()

    assert a1.spec_uuid() == a2.spec_uuid()
    assert dag1.spec_uuid() == dag2.spec_uuid()

    assert a1.spec_uuid() == a3.spec_uuid()
    assert dag1.spec_uuid() != dag3.spec_uuid()


def test_workflow_determinism_8():
    dag1 = FugueWorkflow()
    a1 = dag1.create_data([[0], [0], [1]], "a:int32")
    a1.select("a", "b")
    a1.show()

    dag2 = FugueWorkflow()
    a2 = dag2.create_data([[0], [0], [1]], "a:int32")
    a2.select("a", "b")
    a2.show()

    dag3 = FugueWorkflow()
    a3 = dag3.create_data([[0], [0], [1]], "a:int32")
    a3.select("b", "a")
    a3.show()

    dag4 = FugueWorkflow()
    a4 = dag4.create_data([[0], [0], [1]], "a:int32")
    a4.select("a", "b", distinct=True)
    a4.show()

    assert a1.spec_uuid() == a2.spec_uuid()
    assert dag1.spec_uuid() == dag2.spec_uuid()

    assert a1.spec_uuid() == a3.spec_uuid()
    assert dag1.spec_uuid() != dag3.spec_uuid()

    assert a1.spec_uuid() == a4.spec_uuid()
    assert dag1.spec_uuid() != dag4.spec_uuid()


def mock_tf1(df: List[Dict[str, Any]], v: int = 1) -> Iterable[Dict[str, Any]]:
    for r in df:
        r["b"] = v * len(df)
        yield r



from typing import Any, Dict, Iterable, List, Optional, Tuple

import pandas as pd
import pyarrow as pa
from triad.collections.schema import Schema
from triad.utils.assertion import assert_or_throw
from triad.utils.pandas_like import PD_UTILS

from fugue.dataset.api import (
    as_fugue_dataset,
    as_local,
    as_local_bounded,
    count,
    get_num_partitions,
    is_bounded,
    is_empty,
    is_local,
)
from fugue.exceptions import FugueDataFrameOperationError

from .api import (
    drop_columns,
    get_column_names,
    get_schema,
    head,
    is_df,
    rename,
    select_columns,
)
from .dataframe import DataFrame, LocalBoundedDataFrame, _input_schema


class PandasDataFrame(LocalBoundedDataFrame):
    """DataFrame that wraps pandas DataFrame. Please also read
    |DataFrameTutorial| to understand this Fugue concept

    :param df: 2-dimensional array, iterable of arrays or pandas DataFrame
    :param schema: |SchemaLikeObject|
    :param pandas_df_wrapper: if this is a simple wrapper, default False

    .. admonition:: Examples

        >>> PandasDataFrame([[0,'a'],[1,'b']],"a:int,b:str")
        >>> PandasDataFrame(schema = "a:int,b:int")  # empty dataframe
        >>> PandasDataFrame(pd.DataFrame([[0]],columns=["a"]))
        >>> PandasDataFrame(ArrayDataFrame([[0]],"a:int).as_pandas())

    .. note::

        If ``pandas_df_wrapper`` is True, then the constructor will not do any type
        check otherwise, it will enforce type according to the input schema after
        the construction
    """

    def __init__(  # noqa: C901
        self,
        df: Any = None,
        schema: Any = None,
        pandas_df_wrapper: bool = False,
    ):
        apply_schema = True
        if df is None:
            schema = _input_schema(schema).assert_not_empty()
            df = []
        if isinstance(df, PandasDataFrame):
            # TODO: This is useless if in this way and wrong
            pdf = df.native
            schema = None
        elif isinstance(df, (pd.DataFrame, pd.Series)):
            if isinstance(df, pd.Series):
                df = df.to_frame()
            pdf = df
            schema = None if schema is None else _input_schema(schema)
            if pandas_df_wrapper and schema is not None:
                apply_schema = False
        elif isinstance(df, Iterable):
            schema = _input_schema(schema).assert_not_empty()
            pdf = pd.DataFrame(df, columns=schema.names)
            pdf = PD_UTILS.enforce_type(pdf, schema.pa_schema, null_safe=True)
            if PD_UTILS.empty(pdf):
                for k, v in schema.items():
                    pdf[k] = pdf[k].astype(v.type.to_pandas_dtype())
            apply_schema = False
        else:
            raise ValueError(f"{df} is incompatible with PandasDataFrame")
        if apply_schema:
            pdf, schema = self._apply_schema(pdf, schema)
        super().__init__(schema)
        self._native = pdf

    @property
    def native(self) -> pd.DataFrame:
        """Pandas DataFrame"""
        return self._native

    def native_as_df(self) -> pd.DataFrame:
        return self._native

    @property
    def empty(self) -> bool:
        return self.native.empty

    def peek_array(self) -> List[Any]:
        self.assert_not_empty()
        return self.native.iloc[0].values.tolist()

    def count(self) -> int:
        return self.native.shape[0]

    def as_pandas(self) -> pd.DataFrame:
        return self._native

    def _drop_cols(self, cols: List[str]) -> DataFrame:
        cols = (self.schema - cols).names
        return self._select_cols(cols)

    def _select_cols(self, cols: List[Any]) -> DataFrame:
        schema = self.schema.extract(cols)
        return PandasDataFrame(
            self.native[schema.names], schema, pandas_df_wrapper=True
        )

    def rename(self, columns: Dict[str, str]) -> DataFrame:
        try:
            schema = self.schema.rename(columns)
        except Exception as e:
            raise FugueDataFrameOperationError from e
        df = self.native.rename(columns=columns)
        return PandasDataFrame(df, schema, pandas_df_wrapper=True)

    def alter_columns(self, columns: Any) -> DataFrame:
        new_schema = self._get_altered_schema(columns)
        if new_schema == self.schema:
            return self
        new_pdf = pd.DataFrame(self.native.to_dict(orient="series"))
        for k, v in new_schema.items():
            if not v.type.equals(self.schema[k].type):
                old_type = self.schema[k].type
                new_type = v.type
                # int -> str
                if pa.types.is_integer(old_type) and pa.types.is_string(new_type):
                    series = new_pdf[k]
                    ns = series.isnull()
                    series = series.fillna(0).astype(int).astype(str)
                    new_pdf[k] = series.mask(ns, None)
                # bool -> str
                elif pa.types.is_boolean(old_type) and pa.types.is_string(new_type):
                    series = new_pdf[k]
                    ns = series.isnull()
                    positive = series != 0
                    new_pdf[k] = "False"
                    new_pdf[k] = new_pdf[k].mask(positive, "True").mask(ns, None)
        return PandasDataFrame(new_pdf, new_schema)

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        return list(self.as_array_iterable(columns, type_safe=type_safe))

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        for row in PD_UTILS.as_array_iterable(
            self.native,
            schema=self.schema.pa_schema,
            columns=columns,
            type_safe=type_safe,
        ):
            yield row

    def head(
        self, n: int, columns: Optional[List[str]] = None
    ) -> LocalBoundedDataFrame:
        pdf = self.native if columns is None else self.native[columns]
        schema = self.schema if columns is None else self.schema.extract(columns)
        return PandasDataFrame(pdf.head(n), schema=schema, pandas_df_wrapper=True)

    def _apply_schema(
        self, pdf: pd.DataFrame, schema: Optional[Schema]
    ) -> Tuple[pd.DataFrame, Schema]:
        PD_UTILS.ensure_compatible(pdf)
        if pdf.columns.dtype == "object":  # pdf has named schema
            pschema = _input_schema(pdf)
            if schema is None or pschema == schema:
                return pdf, pschema.assert_not_empty()
            pdf = pdf[schema.assert_not_empty().names]
        else:  # pdf has no named schema
            schema = _input_schema(schema).assert_not_empty()
            assert_or_throw(
                pdf.shape[1] == len(schema),
                lambda: ValueError(
                    f"Pandas datafame column count doesn't match {schema}"
                ),
            )
            pdf.columns = schema.names
        return PD_UTILS.enforce_type(pdf, schema.pa_schema, null_safe=True), schema


@as_local.candidate(lambda df: isinstance(df, pd.DataFrame))
def _pd_as_local(df: pd.DataFrame) -> pd.DataFrame:
    return df


@as_local_bounded.candidate(lambda df: isinstance(df, pd.DataFrame))
def _pd_as_local_bounded(df: pd.DataFrame) -> pd.DataFrame:
    return df


@as_fugue_dataset.candidate(lambda df, **kwargs: isinstance(df, pd.DataFrame))
def _pd_as_fugue_df(df: pd.DataFrame, **kwargs: Any) -> "PandasDataFrame":
    return PandasDataFrame(df, **kwargs)


@is_df.candidate(lambda df: isinstance(df, pd.DataFrame))
def _pd_is_df(df: pd.DataFrame) -> bool:
    return True


@count.candidate(lambda df: isinstance(df, pd.DataFrame))
def _pd_count(df: pd.DataFrame) -> int:
    return df.shape[0]


@is_bounded.candidate(lambda df: isinstance(df, pd.DataFrame))
def _pd_is_bounded(df: pd.DataFrame) -> bool:
    return True


@is_empty.candidate(lambda df: isinstance(df, pd.DataFrame))
def _pd_is_empty(df: pd.DataFrame) -> bool:
    return df.shape[0] == 0


@is_local.candidate(lambda df: isinstance(df, pd.DataFrame))
def _pd_is_local(df: pd.DataFrame) -> bool:
    return True


@get_num_partitions.candidate(lambda df: isinstance(df, pd.DataFrame))
def _get_pandas_num_partitions(df: pd.DataFrame) -> int:
    return 1


@get_column_names.candidate(lambda df: isinstance(df, pd.DataFrame))
def _get_pandas_dataframe_columns(df: pd.DataFrame) -> List[Any]:
    return list(df.columns)


@get_schema.candidate(lambda df: isinstance(df, pd.DataFrame))
def _get_pandas_dataframe_schema(df: pd.DataFrame) -> Schema:
    return Schema(df)


@rename.candidate(lambda df, *args, **kwargs: isinstance(df, pd.DataFrame))
def _rename_pandas_dataframe(
    df: pd.DataFrame, columns: Dict[str, Any], as_fugue: bool = False
) -> Any:
    if len(columns) == 0:
        return df
    _assert_no_missing(df, columns.keys())
    return _adjust_df(df.rename(columns=columns), as_fugue=as_fugue)


@drop_columns.candidate(lambda df, *args, **kwargs: isinstance(df, pd.DataFrame))
def _drop_pd_columns(
    df: pd.DataFrame, columns: List[str], as_fugue: bool = False
) -> Any:
    cols = [x for x in df.columns if x not in columns]
    if len(cols) == 0:
        raise FugueDataFrameOperationError("cannot drop all columns")
    if len(cols) + len(columns) != len(df.columns):
        _assert_no_missing(df, columns)
    return _adjust_df(df[cols], as_fugue=as_fugue)


@select_columns.candidate(lambda df, *args, **kwargs: isinstance(df, pd.DataFrame))
def _select_pd_columns(
    df: pd.DataFrame, columns: List[Any], as_fugue: bool = False
) -> Any:
    if len(columns) == 0:
        raise FugueDataFrameOperationError("must select at least one column")
    _assert_no_missing(df, columns)
    return _adjust_df(df[columns], as_fugue=as_fugue)


@head.candidate(lambda df, *args, **kwargs: isinstance(df, pd.DataFrame))
def _pd_head(
    df: pd.DataFrame,
    n: int,
    columns: Optional[List[str]] = None,
    as_fugue: bool = False,
) -> pd.DataFrame:
    if columns is not None:
        df = df[columns]
    return _adjust_df(df.head(n), as_fugue=as_fugue)


def _adjust_df(res: pd.DataFrame, as_fugue: bool):
    return res if not as_fugue else PandasDataFrame(res)


def _assert_no_missing(df: pd.DataFrame, columns: Iterable[Any]) -> None:
    missing = [x for x in columns if x not in df.columns]
    if len(missing) > 0:
        raise FugueDataFrameOperationError("found nonexistent columns: {missing}")
